"""Test the :mod:`fedora_messaging.api` APIs on a real broker."""

import re
import socket
import time
import uuid
from collections import defaultdict
from unittest import mock
from urllib.parse import quote

import pytest
import pytest_twisted
import treq
from twisted.internet import defer, reactor, task, threads

from fedora_messaging import api, config, exceptions, message
from fedora_messaging.twisted.consumer import _add_timeout

from .utils import RABBITMQ_HOST


HTTP_API = f"http://{RABBITMQ_HOST}:15672/api/"
HTTP_AUTH = ("guest", "guest")


def setup_function(function):
    """Ensure each test starts with a fresh Service and configuration."""
    config.conf = config.LazyConfig()
    config.conf["client_properties"]["app"] = function.__name__
    config.conf["amqp_url"] = f"amqp://{RABBITMQ_HOST}"
    if api._twisted_service:
        pytest_twisted.blockon(api._twisted_service.stopService())
    api._twisted_service = None


@pytest.fixture
def admin_user():
    """
    Fixture that creates a random admin user and deletes the user afterwards.

    Useful if the test wishes to alter permissions to test failure cases. Default
    permissions is complete access to the "/" vhost.

    Returns:
        The username of the new administrator. The password is "guest".
    """
    # Create a user with no permissions
    username = str(uuid.uuid4())
    url = f"{HTTP_API}users/{username}"
    body = {"username": username, "password": "guest", "tags": "administrator"}
    deferred_resp = treq.put(url, json=body, auth=HTTP_AUTH, timeout=3)

    @pytest_twisted.inlineCallbacks
    def cp(resp):
        assert resp.code == 201
        url = f"{HTTP_API}permissions/%2F/{username}"
        body = {"configure": ".*", "write": ".*", "read": ".*"}
        resp = yield treq.put(url, json=body, auth=HTTP_AUTH, timeout=3)
        assert resp.code == 201

    deferred_resp.addCallbacks(cp, cp)
    pytest_twisted.blockon(deferred_resp)
    yield username

    # Cleanup
    deferred_resp = treq.delete(url, auth=HTTP_AUTH, timeout=3)
    pytest_twisted.blockon(deferred_resp)


@pytest_twisted.inlineCallbacks
def get_queue(queue_dict, delay=10):
    """
    Retrieve data about a queue in the broker in the default vhost ("/").

    Args:
        delay: Time to wait before querying the server. It can take some time for
            everything to settle and become available in the HTTP API. 10 seconds
            was arrived at after intermittent failures at lower values.

    Returns:
        dict: A dictionary representation of the queue.
    """
    queues = list(queue_dict.keys())
    name = queues[0]
    # Assert both messages are delivered, no messages are un-acked, and only one
    # message got a positive acknowledgment.
    url = f"{HTTP_API}queues/%2F/{name}"
    server_queue = yield task.deferLater(reactor, delay, treq.get, url, auth=HTTP_AUTH, timeout=3)
    server_queue = yield server_queue.json()
    defer.returnValue(server_queue)


@pytest.fixture
def queue_and_binding():
    queue = str(uuid.uuid4())
    queues = {
        queue: {
            "durable": False,
            "exclusive": False,
            "auto_delete": False,
            "arguments": {"x-expires": 60 * 1000},
        }
    }
    bindings = [{"queue": queue, "exchange": "amq.topic", "routing_keys": ["#"]}]
    yield queues, bindings


@pytest_twisted.inlineCallbacks
def test_twisted_consume_halt_consumer(queue_and_binding):
    """
    Assert raising HaltConsumer works with :func:`fedora_messaging.api.twisted_consume`
    API.
    """
    queues, bindings = queue_and_binding
    msg = message.Message(
        topic="nice.message",
        headers={"niceness": "very"},
        body={"encouragement": "You're doing great!"},
    )
    expected_headers = {
        "fedora_messaging_severity": 20,
        "fedora_messaging_schema": "base.message",
        "priority": 0,
        "niceness": "very",
    }
    messages_received = []

    def callback(message):
        """Count to 3 and quit."""
        messages_received.append(message)
        if len(messages_received) == 3:
            raise exceptions.HaltConsumer()

    consumers = yield api.twisted_consume(callback, bindings, queues)

    # Assert the server reports a consumer
    server_queue = yield get_queue(queues)
    assert server_queue["consumers"] == 1

    for _ in range(0, 3):
        yield threads.deferToThread(api.publish, msg, "amq.topic")

    _add_timeout(consumers[0].result, 10)
    try:
        yield consumers[0].result
    except exceptions.HaltConsumer as e:
        assert len(messages_received) == 3
        assert e.exit_code == 0
        for m in messages_received:
            assert "nice.message" == m.topic
            assert {"encouragement": "You're doing great!"} == m.body
            assert "sent-at" in m._headers
            del m._headers["sent-at"]
            assert expected_headers == m._headers
        server_queue = yield get_queue(queues)
        assert server_queue["consumers"] == 0
    except (defer.TimeoutError, defer.CancelledError):
        yield consumers[0].cancel()
        pytest.fail("Timeout reached without consumer halting!")


@pytest_twisted.inlineCallbacks
def test_twisted_stop_service(queue_and_binding):
    """
    Assert stopping the service, which happens when the reactor shuts down, waits
    for consumers to finish processing and then cancels them before closing the connection.
    """
    queues, bindings = queue_and_binding
    message_received, message_processed = defer.Deferred(), defer.Deferred()

    def callback(message):
        """Callback when the message is received, introduce a delay, then callback again."""
        reactor.callFromThread(message_received.callback, None)
        time.sleep(5)
        reactor.callFromThread(message_processed.callback, None)

    consumers = yield api.twisted_consume(callback, bindings, queues)
    yield threads.deferToThread(api.publish, message.Message(), "amq.topic")

    _add_timeout(consumers[0].result, 10)
    _add_timeout(message_received, 10)
    try:
        yield message_received
    except (defer.TimeoutError, defer.CancelledError):
        yield consumers[0].cancel()
        pytest.fail("Timeout reached without consumer receiving message!")

    assert not message_processed.called
    deferred_stop = api._twisted_service.stopService()

    _add_timeout(message_processed, 10)
    try:
        yield message_processed
        # The request to stop should wait on the message to be processed
        assert deferred_stop.called is False
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Timeout reached without consumer processing message")
    yield deferred_stop


@pytest_twisted.inlineCallbacks
def test_twisted_consume_cancel(queue_and_binding):
    """Assert canceling works with :func:`fedora_messaging.api.twisted_consume` API."""
    queues, bindings = queue_and_binding

    # Assert that the number of consumers we think we started is the number the
    # server things we started. This will fail if other tests don't clean up properly.
    # If it becomes problematic perhaps each test should have a vhost.
    consumers = yield api.twisted_consume(lambda m: m, bindings, queues)
    consumers[0].result.addErrback(pytest.fail)

    server_queue = yield get_queue(queues)
    assert server_queue["consumers"] == 1

    try:
        d = consumers[0].cancel()
        _add_timeout(d, 5)
        yield d

        # Assert the consumer.result deferred has called back when the cancel
        # deferred fires.
        assert consumers[0].result.called

        # Finally make sure the server agrees that the consumer is canceled.
        server_queue = yield get_queue(queues)
        assert server_queue["consumers"] == 0
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Timeout reached without consumer halting!")


@pytest_twisted.inlineCallbacks
def test_twisted_consume_halt_consumer_requeue(queue_and_binding):
    """Assert raising HaltConsumer with requeue=True re-queues the message."""
    queues, bindings = queue_and_binding
    msg = message.Message(
        topic="nice.message",
        headers={"niceness": "very"},
        body={"encouragement": "You're doing great!"},
    )

    def callback(message):
        """Count to 3 and quit."""
        raise exceptions.HaltConsumer(exit_code=1, requeue=True)

    # Assert that the number of consumers we think we started is the number the
    # server things we started. This will fail if other tests don't clean up properly.
    # If it becomes problematic perhaps each test should have a vhost.
    consumers = yield api.twisted_consume(callback, bindings, queues)
    yield threads.deferToThread(api.publish, msg, "amq.topic")

    _add_timeout(consumers[0].result, 10)
    try:
        yield consumers[0].result
    except exceptions.HaltConsumer as e:
        # Assert there are no consumers for the queue, and that there's a ready message
        assert e.exit_code == 1

        server_queue = yield get_queue(queues)
        assert server_queue["consumers"] == 0
        assert server_queue["messages_ready"] == 1
    except (defer.TimeoutError, defer.CancelledError):
        yield consumers[0].cancel()
        pytest.fail("Timeout reached without consumer halting!")


@pytest_twisted.inlineCallbacks
def test_twisted_consume_drop_message(queue_and_binding):
    """Assert raising Drop causes the message to be dropped, but processing continues."""
    queues, bindings = queue_and_binding
    msg = message.Message(
        topic="nice.message",
        headers={"niceness": "very"},
        body={"encouragement": "You're doing great!"},
    )
    dropped_messages = []

    def callback(message):
        """Drop 1 message and then halt on the second message."""
        dropped_messages.append(message)
        if len(dropped_messages) == 2:
            raise exceptions.HaltConsumer()
        raise exceptions.Drop()

    # Assert that the number of consumers we think we started is the number the
    # server things we started. This will fail if other tests don't clean up properly.
    # If it becomes problematic perhaps each test should have a vhost.
    consumers = yield api.twisted_consume(callback, bindings, queues)
    yield threads.deferToThread(api.publish, msg, "amq.topic")
    yield threads.deferToThread(api.publish, msg, "amq.topic")

    _add_timeout(consumers[0].result, 10)
    try:
        yield consumers[0].result
    except exceptions.HaltConsumer:
        # Assert both messages are delivered, no messages are un-acked, and only one
        # message got a positive acknowledgment.
        server_queue = yield get_queue(queues)
        assert server_queue["consumers"] == 0
        assert server_queue["messages"] == 0
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Timeout reached without consumer halting!")
    finally:
        yield consumers[0].cancel()


@pytest_twisted.inlineCallbacks
def test_twisted_consume_nack_message(queue_and_binding):
    """Assert raising Nack causes the message to be replaced in the queue."""
    queues, bindings = queue_and_binding
    msg = message.Message(
        topic="nice.message",
        headers={"niceness": "very"},
        body={"encouragement": "You're doing great!"},
    )
    nacked_messages = []

    def callback(message):
        """Nack the message, then halt."""
        nacked_messages.append(message)
        if len(nacked_messages) == 2:
            raise exceptions.HaltConsumer()
        raise exceptions.Nack()

    # Assert that the number of consumers we think we started is the number the
    # server things we started. This will fail if other tests don't clean up properly.
    # If it becomes problematic perhaps each test should have a vhost.
    consumers = yield api.twisted_consume(callback, bindings, queues)
    yield threads.deferToThread(api.publish, msg, "amq.topic")

    _add_timeout(consumers[0].result, 10)
    try:
        yield consumers[0].result
    except exceptions.HaltConsumer:
        # Assert the message was delivered, redelivered when Nacked, then acked by HaltConsumer
        server_queue = yield get_queue(queues)
        assert server_queue["consumers"] == 0
        assert server_queue["messages"] == 0
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Timeout reached without consumer halting!")
    finally:
        yield consumers[0].cancel()


@pytest_twisted.inlineCallbacks
def test_twisted_consume_general_exception(queue_and_binding):
    """
    Assert if the callback raises an unhandled exception, it is passed on to the
    consumer.result and the message is re-queued.
    """
    queues, bindings = queue_and_binding
    msg = message.Message(
        topic="nice.message",
        headers={"niceness": "very"},
        body={"encouragement": "You're doing great!"},
    )

    def callback(message):
        """An *exceptionally* useless callback"""
        raise Exception("Oh the huge manatee")

    # Assert that the number of consumers we think we started is the number the
    # server things we started. This will fail if other tests don't clean up properly.
    # If it becomes problematic perhaps each test should have a vhost.
    consumers = yield api.twisted_consume(callback, bindings, queues)
    yield threads.deferToThread(api.publish, msg, "amq.topic")

    _add_timeout(consumers[0].result, 10)
    try:
        yield consumers[0].result
        pytest.fail("Expected an exception to be raised.")
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Timeout reached without consumer halting!")
    except Exception as e:
        # Assert the message was delivered and re-queued when the consumer crashed.
        assert e.args[0] == "Oh the huge manatee"
        server_queue = yield get_queue(queues)
        assert server_queue["consumers"] == 0
        assert server_queue["messages"] == 1
    finally:
        yield consumers[0].cancel()


@pytest_twisted.inlineCallbacks
def test_twisted_consume_connection_reset(queue_and_binding):
    """
    Assert consuming works across connections and handles connection resets.

    This test sets up a queue, publishes 2 messages to itself, then kills all active
    connections on the broker. It then sends a third message and asserts the consumer
    gets it.
    """
    queues, bindings = queue_and_binding
    msg = message.Message(
        topic="nice.message",
        headers={"niceness": "very"},
        body={"encouragement": "You're doing great!"},
    )
    messages_received = []
    two_received = defer.Deferred()  # Fired by the callback on 2 messages

    def callback(message):
        """Count to, 2, fire a deferred, then count to 3 and quit."""
        messages_received.append(message)
        if len(messages_received) == 2:
            two_received.callback(None)
        if len(messages_received) == 3:
            raise exceptions.HaltConsumer()

    consumers = yield api.twisted_consume(callback, bindings, queues)

    # Wait for two messages to get through, kill the connection, and then send
    # the third and wait for the consumer to finish
    yield threads.deferToThread(api.publish, msg, "amq.topic")
    yield threads.deferToThread(api.publish, msg, "amq.topic")
    _add_timeout(two_received, 10)
    try:
        yield two_received
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Timeout reached without receiving first two messages")

    for _ in range(10):
        conns = yield task.deferLater(
            reactor, 1, treq.get, HTTP_API + "connections", auth=HTTP_AUTH, timeout=3
        )
        conns = yield conns.json()
        this_conn = [
            c
            for c in conns
            if "app" in c["client_properties"]
            and c["client_properties"]["app"] == "test_twisted_consume_connection_reset"
        ]
        if this_conn:
            cname = quote(this_conn[0]["name"])
            yield treq.delete(HTTP_API + "connections/" + cname, auth=HTTP_AUTH, timeout=3)
            break
    else:
        pytest.fail("Unable to find and kill connection!")

    # The consumer should receive this third message after restarting its connection
    # and then it should exit gracefully.
    yield threads.deferToThread(api.publish, msg, "amq.topic")
    _add_timeout(consumers[0].result, 10)
    try:
        yield consumers[0].result
    except exceptions.HaltConsumer:
        assert len(messages_received) == 3
    except (defer.TimeoutError, defer.CancelledError):
        yield consumers[0].cancel()
        pytest.fail("Timeout reached without consumer halting!")


@pytest_twisted.inlineCallbacks
def test_twisted_consume_serverside_cancel(queue_and_binding):
    """
    Assert the consumer halts and ``consumer.result`` errbacks when the server
    explicitly cancels the consumer (by deleting the queue).
    """
    queues, bindings = queue_and_binding

    consumers = yield api.twisted_consume(lambda x: x, bindings, queues)

    # Delete the queue and assert the consumer errbacks
    url = f"{HTTP_API}queues/%2F/{next(iter(queues.keys()))}"
    yield treq.delete(url, auth=HTTP_AUTH, timeout=3)

    _add_timeout(consumers[0].result, 10)
    try:
        yield consumers[0].result
        pytest.fail("Consumer did not errback!")
    except exceptions.ConsumerCanceled:
        pass
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Timeout reached without consumer calling its errback!")


@pytest_twisted.inlineCallbacks
def test_no_vhost_permissions(admin_user, queue_and_binding):
    """Assert a hint is given if the user doesn't have any access to the vhost"""
    url = f"{HTTP_API}permissions/%2F/{admin_user}"
    resp = yield treq.delete(url, auth=HTTP_AUTH, timeout=3)
    assert resp.code == 204

    queues, _ = queue_and_binding

    amqp_url = f"amqp://{admin_user}:guest@{RABBITMQ_HOST}:5672/%2F"
    with mock.patch.dict(config.conf, {"amqp_url": amqp_url}):
        try:
            yield api.twisted_consume(lambda x: x, [], queues)
        except exceptions.ConnectionException as e:
            assert e.reason == (
                "The TCP connection appears to have started, but the TLS or AMQP "
                "handshake with the broker failed; check your connection and "
                "authentication parameters and ensure your user has permission "
                "to access the vhost"
            )


@pytest_twisted.inlineCallbacks
def test_no_read_permissions_queue_read_failure_pika1(admin_user, queue_and_binding):
    """
    Assert an errback occurs when unable to read from the queue due to
    permissions. This is a bit weird because the consumer has permission to
    register itself, but not to actually read from the queue so the result is
    what errors back.
    """
    queues, bindings = queue_and_binding
    url = f"{HTTP_API}permissions/%2F/{admin_user}"
    body = {"configure": ".*", "write": ".*", "read": ""}
    resp = yield treq.put(url, json=body, auth=HTTP_AUTH, timeout=3)
    assert resp.code == 204

    amqp_url = f"amqp://{admin_user}:guest@{RABBITMQ_HOST}:5672/%2F"
    with mock.patch.dict(config.conf, {"amqp_url": amqp_url}):
        try:
            consumers = api.twisted_consume(lambda x: x, [], queues)
            _add_timeout(consumers, 5)
            yield consumers
            pytest.fail("Call failed to raise an exception")
        except exceptions.PermissionException as e:
            error_match = re.match(
                r"ACCESS_REFUSED - (read )?access to queue '[\w-]+' in vhost '/' refused "
                r"for user '[\w-]+'",
                e.reason,
            )
            assert error_match is not None
        except (defer.TimeoutError, defer.CancelledError):
            pytest.fail("Timeout reached without consumer calling its errback!")


@pytest_twisted.inlineCallbacks
def test_no_read_permissions_bind_failure(admin_user, queue_and_binding):
    """Assert the call to twisted_consume errbacks on read permissions errors on binding."""
    queues, bindings = queue_and_binding
    url = f"{HTTP_API}permissions/%2F/{admin_user}"
    body = {"configure": ".*", "write": ".*", "read": ""}
    resp = yield treq.put(url, json=body, auth=HTTP_AUTH, timeout=3)
    assert resp.code == 204

    amqp_url = f"amqp://{admin_user}:guest@{RABBITMQ_HOST}:5672/%2F"
    try:
        with mock.patch.dict(config.conf, {"amqp_url": amqp_url}):
            yield api.twisted_consume(lambda x: x, bindings, queues)
        pytest.fail("Call failed to raise an exception")
    except exceptions.BadDeclaration as e:
        assert e.reason.args[0] == 403
        error_match = re.match(
            r"ACCESS_REFUSED - (read )?access to exchange 'amq\.topic' in vhost '/' refused "
            r"for user '[\w-]+'",
            e.reason.args[1],
        )
        assert error_match is not None


@pytest_twisted.inlineCallbacks
def test_no_write_permissions(admin_user, queue_and_binding):
    """Assert the call to twisted_consume errbacks on write permissions errors."""
    queues, bindings = queue_and_binding
    url = f"{HTTP_API}permissions/%2F/{admin_user}"
    body = {"configure": ".*", "write": "", "read": ".*"}
    resp = yield treq.put(url, json=body, auth=HTTP_AUTH, timeout=3)
    assert resp.code == 204

    amqp_url = f"amqp://{admin_user}:guest@{RABBITMQ_HOST}:5672/%2F"
    try:
        with mock.patch.dict(config.conf, {"amqp_url": amqp_url}):
            yield api.twisted_consume(lambda x: x, bindings, queues)
        pytest.fail("Call failed to raise an exception")
    except exceptions.BadDeclaration as e:
        assert e.reason.args[0] == 403
        error_match = re.match(
            r"ACCESS_REFUSED - (write )?access to queue '[\w-]+' in vhost '/' refused "
            r"for user '[\w-]+'",
            e.reason.args[1],
        )
        assert error_match is not None


@pytest_twisted.inlineCallbacks
def test_twisted_consume_invalid_message(queue_and_binding):
    """Assert messages that fail validation are nacked."""
    queues, bindings = queue_and_binding
    bindings[0]["routing_keys"] = ["test_twisted_invalid_message"]

    consumers = yield api.twisted_consume(
        lambda x: pytest.fail("Message should be nacked"), bindings, queues
    )

    url = HTTP_API + "exchanges/%2F/amq.topic/publish"
    body = {
        "properties": {},
        "routing_key": "test_twisted_invalid_message",
        "payload": "not json",
        "payload_encoding": "string",
    }
    response = yield treq.post(url, json=body, auth=HTTP_AUTH, timeout=3)
    response = yield response.json()
    assert response["routed"] is True
    yield consumers[0].cancel()


@pytest_twisted.inlineCallbacks
def test_twisted_consume_update_callback(queue_and_binding):
    """Assert a second call to consume updates an existing callback."""
    queues, bindings = queue_and_binding

    callback1 = defer.Deferred()
    callback2 = defer.Deferred()

    consumers1 = yield api.twisted_consume(
        lambda m: reactor.callFromThread(callback1.callback, m), bindings, queues
    )
    yield threads.deferToThread(api.publish, message.Message(), "amq.topic")
    _add_timeout(callback1, 10)
    try:
        yield callback1
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Never received message for initial callback")

    consumers2 = yield api.twisted_consume(
        lambda m: reactor.callFromThread(callback2.callback, m), bindings, queues
    )
    yield threads.deferToThread(api.publish, message.Message(), "amq.topic")
    _add_timeout(callback2, 10)
    try:
        yield callback2
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Never received message for updated callback")

    assert consumers1[0]._tag == consumers2[0]._tag

    yield consumers2[0].cancel()


@pytest_twisted.inlineCallbacks
def test_publish_channel_error(queue_and_binding):
    """Assert publishing recovers from a channel error."""
    queues, bindings = queue_and_binding
    storage = defaultdict(int)

    def counting_callback(message):
        storage[message.topic] += 1
        if storage[message.topic] == 2:
            raise exceptions.HaltConsumer()

    @pytest_twisted.inlineCallbacks
    def delayed_publish():
        """Publish, break the channel, and publish again."""
        yield threads.deferToThread(api.publish, message.Message(), "amq.topic")
        protocol = yield api._twisted_service._service.factory.when_connected()
        yield protocol._publish_channel.close()
        yield threads.deferToThread(api.publish, message.Message(), "amq.topic")

    reactor.callLater(5, delayed_publish)

    deferred_consume = threads.deferToThread(api.consume, counting_callback, bindings, queues)
    deferred_consume.addTimeout(30, reactor)
    try:
        yield deferred_consume
        pytest.fail("consume should have raised an exception")
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Consumer did not receive both messages")
    except exceptions.HaltConsumer as e:
        assert 0 == e.exit_code


@pytest_twisted.inlineCallbacks
def test_protocol_publish_connection_error(queue_and_binding):
    """Assert individual protocols raise connection exceptions if closed."""
    api._init_twisted_service()
    protocol = yield api._twisted_service._service.factory.when_connected()
    yield protocol.close()
    try:
        yield protocol.publish(message.Message(), "amq.topic")
        pytest.fail("Expected a ConnectionException")
    except exceptions.ConnectionException:
        pass


@pytest_twisted.inlineCallbacks
def test_protocol_publish_forbidden(admin_user):
    """Assert individual protocols raise exceptions when the topic is not allowed."""
    url = f"{HTTP_API}topic-permissions/%2F/{admin_user}"
    body = {"exchange": "amq.topic", "write": "^allowed$", "read": ".*"}
    resp = yield treq.put(url, json=body, auth=HTTP_AUTH, timeout=3)
    assert resp.code == 201

    amqp_url = f"amqp://{admin_user}:guest@{RABBITMQ_HOST}:5672/%2F"
    with mock.patch.dict(config.conf, {"amqp_url": amqp_url}):
        try:
            api._init_twisted_service()
            protocol = yield api._twisted_service._service.factory.when_connected()
            d = protocol.publish(message.Message(topic="not-allowed"), "amq.topic")
            d.addTimeout(5, reactor)
            yield d
            pytest.fail("Publish failed to raise an exception")
        except (defer.TimeoutError, defer.CancelledError):
            pytest.fail("Publishing hit the timeout, probably stuck in a retry loop")
        except exceptions.PublishForbidden as e:
            assert e.reason.args[0] == 403
            error_match = re.match(
                r"ACCESS_REFUSED - (write )?access to topic 'not-allowed' in "
                r"exchange 'amq.topic' in vhost '/' refused for user '[\w-]+'",
                e.reason.args[1],
            )
            assert error_match is not None
        # Now try on an allowed topic
        yield protocol.publish(message.Message(topic="allowed"), "amq.topic")


@pytest_twisted.inlineCallbacks
def test_protocol_publish_forbidden_in_vhost(admin_user):
    """Assert individual protocols raise a forbidden exception immediately when
    the user is not allowed to publish against the virtual host."""
    url = f"{HTTP_API}permissions/%2F/{admin_user}"
    body = {"configure": ".*", "write": "", "read": ".*"}
    resp = yield treq.put(url, json=body, auth=HTTP_AUTH, timeout=3)
    assert resp.code == 204

    amqp_url = f"amqp://{admin_user}:guest@{RABBITMQ_HOST}:5672/%2F"
    with mock.patch.dict(config.conf, {"amqp_url": amqp_url}):
        try:
            api._init_twisted_service()
            protocol = yield api._twisted_service._service.factory.when_connected()
            d = protocol.publish(message.Message(topic="not-allowed"), "amq.topic")
            d.addTimeout(5, reactor)
            yield d
            pytest.fail("Publish failed to raise an exception")
        except (defer.TimeoutError, defer.CancelledError):
            pytest.fail("Publishing hit the timeout, probably stuck in a retry loop")
        except exceptions.PublishForbidden as e:
            assert e.reason.args[0] == 403
            error_match = re.match(
                r"ACCESS_REFUSED - (write )?access to exchange 'amq.topic' in "
                r"vhost '/' refused for user '[\w-]+'",
                e.reason.args[1],
            )
            assert error_match is not None


@pytest_twisted.inlineCallbacks
def test_publish_from_callback(queue_and_binding):
    """Assert a consumer can publish."""
    queues, bindings = queue_and_binding
    received = []

    def publishing_callback(msg):
        received.append(msg)
        if msg.topic == "repeat":
            raise exceptions.HaltConsumer()
        new_msg = message.Message(topic="repeat", body={"received": msg.body})
        threads.blockingCallFromThread(reactor, api.twisted_publish, new_msg)

    @pytest_twisted.inlineCallbacks
    def delayed_publish():
        """Publish, break the channel, and publish again."""
        yield threads.deferToThread(api.publish, message.Message(topic="source"), "amq.topic")
        yield api._twisted_service._service.factory.when_connected()

    reactor.callLater(5, delayed_publish)

    deferred_consume = threads.deferToThread(api.consume, publishing_callback, bindings, queues)
    deferred_consume.addTimeout(30, reactor)
    try:
        yield deferred_consume
        pytest.fail("consume should have raised an exception")
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Consumer did not receive both messages")
    except exceptions.HaltConsumer as e:
        assert 0 == e.exit_code
    assert [m.topic for m in received] == ["source", "repeat"]


@pytest_twisted.inlineCallbacks
def test_pub_sub_default_settings(queue_and_binding):
    """
    Assert publishing and subscribing works with the default configuration.

    This should work because the publisher uses the 'amq.topic' exchange by
    default and the consumer also uses the 'amq.topic' exchange with its
    auto-named queue and a default subscription key of '#'.
    """
    queues, bindings = queue_and_binding
    storage = defaultdict(int)

    # Consumer setup
    def counting_callback(message):
        storage[message.topic] += 1
        if storage[message.topic] == 3:
            raise exceptions.HaltConsumer()

    deferred_consume = threads.deferToThread(api.consume, counting_callback, bindings, queues)

    @pytest_twisted.inlineCallbacks
    def delayed_publish():
        """Give the consumer time to setup."""
        msg = message.Message(
            topic="nice.message",
            headers={"niceness": "very"},
            body={"encouragement": "You're doing great!"},
        )
        for _ in range(0, 3):
            try:
                yield threads.deferToThread(api.publish, msg, "amq.topic")
            except exceptions.ConnectionException:
                pytest.fail("Failed to publish message, is the broker running?")

    reactor.callLater(5, delayed_publish)
    deferred_consume.addTimeout(30, reactor)
    try:
        yield deferred_consume
        pytest.fail("consume should have raised an exception")
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Timeout reached without consumer halting!")
    except exceptions.HaltConsumer as e:
        assert 0 == e.exit_code


@pytest.fixture
def bad_amqp_url():
    amqp_url_backup = config.conf["amqp_url"]
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", 0))
    config.conf["amqp_url"] = f"amqp://{RABBITMQ_HOST}:{sock.getsockname()[1]}/"
    yield
    config.conf["amqp_url"] = amqp_url_backup
    sock.close()
    # The cached service is borked, reset it.
    api._twisted_service = None


@pytest_twisted.inlineCallbacks
def test_pub_timeout(bad_amqp_url):
    """Assert PublishTimeout is raised if a connection just hangs."""
    try:
        yield threads.deferToThread(api.publish, api.Message(), 5)
    except Exception as e:
        if not isinstance(e, exceptions.PublishTimeout):
            pytest.fail(f"Expected a timeout exception, not {e}")
    # Ensure the deferred has been renewed
    assert api._twisted_service._service.factory.when_connected().called is False



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import json
from unittest import mock

import pika
import pytest
from twisted.internet import defer

from fedora_messaging import config
from fedora_messaging.exceptions import (
    ConnectionException,
    NoFreeChannels,
    PublishForbidden,
    PublishReturned,
)
from fedora_messaging.message import INFO, Message
from fedora_messaging.twisted.protocol import Consumer, FedoraMessagingProtocolV2

from .utils import MockProtocol


try:
    import pytest_twisted  # noqa: F401
except ImportError:
    pytest.skip("pytest-twisted is missing, skipping tests", allow_module_level=True)


class TestProtocol:
    def setup_method(self, method):
        self.protocol = MockProtocol(None)
        self.factory = mock.Mock()
        self.protocol.factory = self.factory

    def test_allocate_channel_no_more_channels(self):
        """Assert a pika NoFreeChannels exception turns into a fedora_messaging exception."""
        self.protocol.channel = mock.Mock(
            name="channel", side_effect=pika.exceptions.NoFreeChannels()
        )

        d = self.protocol._allocate_channel()
        d.addCallback(pytest.fail, "Expected a NoFreeChannels exception")
        d.addErrback(lambda f: f.trap(NoFreeChannels))
        return d

    def test_allocate_channel_connection_exception(self):
        """
        Assert a pika ConnectionWrongStateError exception turns into a fedora_messaging exception.
        """
        self.protocol.channel = mock.Mock(
            name="channel", side_effect=pika.exceptions.ConnectionWrongStateError()
        )

        d = self.protocol._allocate_channel()
        d.addCallback(pytest.fail, "Expected a ConnectionException exception")
        d.addErrback(lambda f: f.trap(ConnectionException))
        return d

    @mock.patch("fedora_messaging.twisted.consumer.uuid.uuid4", mock.Mock(return_value="tag1"))
    def test_consume_not_running(self):
        """Assert calling consume results in resumeProducing being invoked."""
        func = mock.Mock()

        def _check(consumer):
            consumer._channel.basic_qos.assert_called_with(
                prefetch_count=config.conf["qos"]["prefetch_count"],
                prefetch_size=config.conf["qos"]["prefetch_size"],
            )
            consumer._channel.basic_consume.assert_called_once_with(
                queue="my_queue", consumer_tag="tag1"
            )
            assert self.protocol._consumers["my_queue"] == consumer

        d = self.protocol.consume(func, "my_queue")
        d.addCallback(_check)

        return d

    def test_consume_twice(self):
        """Assert calling consume on the same queue updates the callback."""

        def cb1():
            pass

        def cb2():
            pass

        def _check(_):
            assert 1 == len(self.protocol._consumers)
            assert cb2 == self.protocol._consumers["my_queue"].callback

        d = self.protocol.consume(cb1, "my_queue")
        d.addCallback(lambda _: self.protocol.consume(cb2, "my_queue"))
        d.addCallback(_check)

        return d

    @mock.patch("fedora_messaging.twisted.consumer.uuid.uuid4", mock.Mock(return_value="tag1"))
    def test_consume_running(self):
        """Assert when running, consume sets up the AMQP consumer"""
        func = mock.Mock()

        def _check(consumer):
            consumer._channel.basic_qos.assert_called_with(
                prefetch_count=config.conf["qos"]["prefetch_count"],
                prefetch_size=config.conf["qos"]["prefetch_size"],
            )
            consumer._channel.basic_consume.assert_called_once_with(
                queue="my_queue", consumer_tag="tag1"
            )
            assert self.protocol._consumers["my_queue"] == consumer

        d = self.protocol.consume(func, "my_queue")
        d.addCallback(_check)

        return d

    def test_cancel(self):
        """Assert a consumer is removed from the consumer list and canceled."""

        def cb():
            pass

        def _check(_):
            assert 1 == len(self.protocol._consumers)

        d = self.protocol.consume(cb, "my_queue")
        d.addCallback(lambda _: self.protocol.consume(cb, "my_queue2"))
        d.addCallback(lambda consumer: consumer.cancel())
        d.addCallback(_check)

        return d

    def test_forget_no_consumer(self):
        """Assert forgetting a non-existent consumer just returns None."""
        result = self.protocol._forget_consumer("my_invalid_queue")
        assert result is None

    def test_connection_ready(self):
        # Check the ready Deferred.
        def _check(_):
            self.protocol.channel.assert_called_once_with()

        d = self.protocol.ready
        d.addCallback(_check)

        self.protocol._on_connection_ready(None)

        return d

    def test_publish(self):
        # Check the publish method.
        body = {"bodykey": "bodyvalue"}
        headers = {"headerkey": "headervalue"}
        message = Message(body, headers, "testing.topic")
        d = self.protocol.publish(message, "test-exchange")

        def _check(_):
            self.protocol._channel.basic_publish.assert_called_once()
            args = self.protocol._channel.basic_publish.call_args_list[0][1]
            assert args["exchange"] == "test-exchange"
            assert args["routing_key"] == b"testing.topic"
            assert args["body"] == json.dumps(body).encode("utf-8")
            props = args["properties"]
            assert props.headers["fedora_messaging_schema"] == "base.message"
            assert props.headers["fedora_messaging_severity"] == INFO
            assert props.headers["headerkey"] == "headervalue"
            assert "sent-at" in props.headers
            assert props.content_encoding == "utf-8"
            assert props.content_type == "application/json"
            assert props.delivery_mode == 2

        d.addCallback(_check)
        return d

    def test_publish_forbidden(self):
        # Check the publish method when publishing is forbidden.
        body = {"bodykey": "bodyvalue"}
        headers = {"headerkey": "headervalue"}
        message = Message(body, headers, "testing.topic")
        self.protocol._channel.basic_publish.side_effect = pika.exceptions.ChannelClosed(
            403, "Test forbidden message"
        )
        d = self.protocol.publish(message, "test-exchange")

        d.addCallback(pytest.fail, "Expected a PublishForbidden exception")
        d.addErrback(lambda f: f.trap(PublishForbidden))

        def _check(_):
            # Make sure the channel will be re-allocated
            assert self.protocol._publish_channel is None

        d.addBoth(_check)

        return d

    def test_publish_access_denied(self):
        body = {"bodykey": "bodyvalue"}
        headers = {"headerkey": "headervalue"}
        message = Message(body, headers, "testing.topic")
        self.protocol._channel.basic_publish.side_effect = (
            pika.exceptions.ProbableAccessDeniedError(403, "Test access denied message")
        )
        d = self.protocol.publish(message, "test-exchange")

        d.addCallback(pytest.fail, "Expected a PublishForbidden exception")
        d.addErrback(lambda f: f.trap(PublishForbidden))

        return d

    def test_publish_returned(self):
        body = {"bodykey": "bodyvalue"}
        headers = {"headerkey": "headervalue"}
        message = Message(body, headers, "testing.topic")
        self.protocol._channel.basic_publish.side_effect = pika.exceptions.NackError([])
        d = self.protocol.publish(message, "test-exchange")

        d.addCallback(pytest.fail, "Expected a PublishReturned exception")
        d.addErrback(lambda f: f.trap(PublishReturned))

        return d

    def test_publish_connection_closed(self):
        body = {"bodykey": "bodyvalue"}
        headers = {"headerkey": "headervalue"}
        message = Message(body, headers, "testing.topic")
        self.protocol._channel.basic_publish.side_effect = pika.exceptions.ConnectionClosed(
            42, "testing"
        )
        d = self.protocol.publish(message, "test-exchange")

        d.addCallback(pytest.fail, "Expected a ConnectionException exception")
        d.addErrback(lambda f: f.trap(ConnectionException))

        return d

    def test_consume_connection_exception(self):
        """If consuming fails due to a non-permission error, a ConnectionException happens."""
        proto = FedoraMessagingProtocolV2(None)
        mock_channel = mock.Mock()
        mock_channel.basic_consume.side_effect = pika.exceptions.ChannelClosed(400, "Bad Request!")
        deferred_channel = defer.succeed(mock_channel)
        proto._allocate_channel = mock.Mock(return_value=deferred_channel)

        def check(failure):
            assert isinstance(failure.value, ConnectionException)

        d = proto.consume(lambda x: x, "test_queue")
        d.addBoth(check)
        return d

    def test_consume_existing_consumer(self):
        """Consuming should re-use an existing consumer if possible."""
        proto = FedoraMessagingProtocolV2(None)
        proto._allocate_channel = mock.Mock()

        new_callback = mock.Mock(name="new_callback")

        # Prepare the existing consumer
        existing_callback = mock.Mock(name="existing_callback")
        consumer = Consumer(queue="test_queue", callback=existing_callback)
        proto._consumers["test_queue"] = consumer

        def check(result):
            assert consumer == result
            assert consumer.callback == new_callback
            proto._allocate_channel.assert_not_called()

        d = proto.consume(new_callback, "test_queue", consumer)
        d.addBoth(check)
        return d

    def test_consume_provided_consumer(self):
        """The consume() method must handle being passed a consumer."""
        proto = FedoraMessagingProtocolV2(None)
        mock_channel = mock.Mock(name="mock_channel")
        deferred_channel = defer.succeed(mock_channel)
        proto._allocate_channel = mock.Mock(return_value=deferred_channel)
        # basic_consume() must return a tuple
        mock_channel.basic_consume.return_value = (mock.Mock(), mock.Mock())
        # Prepare the consumer
        callback = mock.Mock()
        consumer = Consumer(queue="queue_orig", callback=callback)
        # Don't go into the read loop
        consumer._read = mock.Mock(return_value=defer.succeed(None))

        def check(result):
            assert consumer == result
            assert consumer.queue == "queue_new"
            assert consumer._protocol == proto
            assert consumer._channel == mock_channel

        d = proto.consume(callback, "queue_new", consumer)
        d.addBoth(check)
        return d



"""Exceptions raised by Fedora Messaging."""

import jsonschema


class BaseException(Exception):
    """The base class for all exceptions raised by fedora_messaging."""


class NoFreeChannels(BaseException):
    """Raised when a connection has reached its channel limit"""


class PermissionException(BaseException):
    """
    Generic permissions exception.

    Args:
        obj_type (str): The type of object being accessed that caused the
            permission error. May be None if the cause is unknown.
        description (object): The description of the object, if any. May be None.
        reason (str): The reason the server gave for the permission error, if
            any. If no reason is supplied by the server, this should be the best
            guess for what caused the error.
    """

    def __init__(self, obj_type=None, description=None, reason=None):
        self.obj_type = obj_type
        self.description = description
        self.reason = reason

    def __str__(self):
        return self.description

    def __repr__(self):
        return (
            f"PermissionException(obj_type={self.obj_type}, description={self.description}, "
            f"reason={self.reason})"
        )


class BadDeclaration(PermissionException):
    """
    Raised when declaring an object in AMQP fails.

    Args:
        obj_type (str): The type of object being declared. One of "binding",
            "queue", or "exchange".
        description (dict): The description of the object.
        reason (str): The reason the server gave for rejecting the declaration.
    """

    def __str__(self):
        return (
            f"Unable to declare the {self.obj_type} object ({self.description}) "
            f"because {self.reason}"
        )

    def __repr__(self):
        return (
            f"BadDeclaration(obj_type={self.obj_type}, description={self.description}, "
            f"reason={self.reason})"
        )


class ConfigurationException(BaseException):
    """
    Raised when there's an invalid configuration setting

    Args:
        message (str): A detailed description of the configuration problem
                       which is presented to the user.
    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "Configuration error: " + self.message


class PublishException(BaseException):
    """Base class for exceptions related to publishing."""

    def __init__(self, reason=None, **kwargs):
        super().__init__(**kwargs)
        self.reason = reason

    def __str__(self):
        return str(self.reason)


class PublishReturned(PublishException):
    """
    Raised when the broker rejects and returns the message to the publisher.

    You may handle this exception by logging it and resending or discarding the
    message.
    """


class PublishTimeout(PublishException):
    """
    Raised when the message could not be published in the given timeout.

    This means the message was either never delivered to the broker, or that it
    was delivered, but never acknowledged by the broker.
    """


class PublishForbidden(PublishException):
    """
    Raised when the broker rejects the message due to permission errors.

    You may handle this exception by logging it and discarding the message,
    as it is likely a permanent error.
    """


class ConnectionException(BaseException):
    """
    Raised if a general connection error occurred.

    You may handle this exception by logging it and resending or discarding the
    message.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.reason = kwargs.get("reason")


class ConsumeException(BaseException):
    """Base class for exceptions related to consuming."""


class ConsumerCanceled(ConsumeException):
    """
    Raised when the server has canceled the consumer.

    This can happen when the queue the consumer is subscribed to is deleted,
    or when the node the queue is located on fails.
    """


class Nack(ConsumeException):
    """
    Consumer callbacks should raise this to indicate they wish the message they
    are currently processing to be re-queued.
    """


class Drop(ConsumeException):
    """
    Consumer callbacks should raise this to indicate they wish the message they
    are currently processing to be dropped.
    """


class HaltConsumer(ConsumeException):
    """
    Consumer callbacks should raise this exception if they wish the consumer to
    be shut down.

    Args:
        exit_code (int): The exit code to use when halting.
        reason (str): A reason for halting, presented to the user.
        requeue (bool): If true, the message is re-queued for later processing.
    """

    def __init__(self, exit_code=0, reason=None, requeue=False, **kwargs):
        super().__init__(**kwargs)
        self.exit_code = exit_code
        self.reason = reason
        self.requeue = requeue


class ValidationError(BaseException):
    """
    This error is raised when a message fails validation with its JSON schema

    This exception can be raised on an incoming or outgoing message. No need to
    catch this exception when publishing, it should warn you during development
    and testing that you're trying to publish a message with a different
    format, and that you should either fix it or update the schema.
    """

    @property
    def summary(self):
        """A short summary of the error."""
        original_exception = self.args[0]
        if isinstance(original_exception, jsonschema.exceptions.ValidationError):
            return original_exception.message
        return str(original_exception)



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import os
from unittest import mock

import pika
import pytest
from pika import URLParameters
from twisted.application.internet import SSLClient, TCPClient
from twisted.internet import ssl as twisted_ssl

from fedora_messaging import config, exceptions
from fedora_messaging.twisted.factory import FedoraMessagingFactoryV2
from fedora_messaging.twisted.service import (
    _configure_tls_parameters,
    _ssl_context_factory,
    FedoraMessagingServiceV2,
    SSLOptions,
)


class TestService:
    def test_init(self):
        service = FedoraMessagingServiceV2("amqp://example.com:4242")
        assert isinstance(service._parameters, pika.URLParameters)
        assert service._parameters.host == "example.com"
        assert service._parameters.port == 4242
        assert getattr(service._parameters, "ssl", False) is False
        assert service._parameters.client_properties == config.conf["client_properties"]
        assert isinstance(service._service.factory, FedoraMessagingFactoryV2)
        assert len(service.services) == 1
        assert service.services[0] is service._service
        assert service._service.parent is service
        assert isinstance(service._service, TCPClient)

    def test_init_tls(self):
        """Assert creating the service with an amqps URL configures TLS."""
        service = FedoraMessagingServiceV2("amqps://")

        assert isinstance(service._parameters, pika.URLParameters)
        assert service._parameters.ssl_options is not None
        assert len(service.services) == 1
        assert isinstance(service.services[0], SSLClient)

    def test_init_client_props_override(self):
        service = FedoraMessagingServiceV2("amqp://?client_properties={'foo':'bar'}")
        assert service._parameters.client_properties == {"foo": "bar"}

    def test_stopService(self):
        service = FedoraMessagingServiceV2("amqps://")
        service._service = mock.Mock()
        service.stopService()
        service._service.factory.stopTrying.assert_called_once()
        service._service.factory.stopFactory.assert_called_once()


class ConfigureTlsParameters:
    """Tests for :func:`fedora_messaging._session._configure_tls_parameters`"""

    def test_key_and_cert(self, fixtures_dir):
        """Assert configuring a cert and key results in a TLS connection with new pika versions."""
        params = URLParameters("amqps://myhost")
        tls_conf = {
            "keyfile": os.path.join(fixtures_dir, "key.pem"),
            "certfile": os.path.join(fixtures_dir, "cert.pem"),
            "ca_cert": os.path.join(fixtures_dir, "ca_bundle.pem"),
        }

        with mock.patch.dict(config.conf, {"tls": tls_conf}):
            _configure_tls_parameters(params)

        assert isinstance(params.ssl_options, SSLOptions)

    def test__invalid_key(self, fixtures_dir):
        """Assert a ConfigurationException is raised when the key can't be opened."""
        params = URLParameters("amqps://myhost")
        tls_conf = {
            "keyfile": os.path.join(fixtures_dir, "invalid_key.pem"),
            "certfile": os.path.join(fixtures_dir, "cert.pem"),
            "ca_cert": os.path.join(fixtures_dir, "ca_bundle.pem"),
        }

        with mock.patch.dict(config.conf, {"tls": tls_conf}):
            with pytest.raises(exceptions.ConfigurationException):
                _configure_tls_parameters(params)

        assert isinstance(params.ssl_options, SSLOptions)

    def test_invalid_ca_cert(self, fixtures_dir):
        """Assert a ConfigurationException is raised when the CA can't be opened."""
        params = URLParameters("amqps://myhost")
        tls_conf = {
            "keyfile": os.path.join(fixtures_dir, "key.pem"),
            "certfile": os.path.join(fixtures_dir, "cert.pem"),
            "ca_cert": os.path.join(fixtures_dir, "invalid_ca.pem"),
        }

        with mock.patch.dict(config.conf, {"tls": tls_conf}):
            with pytest.raises(exceptions.ConfigurationException):
                _configure_tls_parameters(params)

        assert isinstance(params.ssl_options, SSLOptions)


class TestSslContextFactory:
    @mock.patch("fedora_messaging.twisted.service.twisted_ssl.Certificate.loadPEM")
    @mock.patch("fedora_messaging.twisted.service.twisted_ssl.optionsForClientTLS")
    def test_no_key(self, mock_opts, mock_load_pem, fixtures_dir):
        """Assert if there's no client key, the factory isn't configured to use it."""
        tls_conf = {
            "keyfile": None,
            "certfile": os.path.join(fixtures_dir, "cert.pem"),
            "ca_cert": os.path.join(fixtures_dir, "ca_bundle.pem"),
        }

        with mock.patch.dict(config.conf, {"tls": tls_conf}):
            factory = _ssl_context_factory(mock.Mock(host="dummy_host"))

        mock_opts.assert_called_once_with(
            "dummy_host",
            trustRoot=mock_load_pem.return_value,
            clientCertificate=None,
            extraCertificateOptions={"raiseMinimumTo": twisted_ssl.TLSVersion.TLSv1_2},
        )
        assert factory == mock_opts.return_value

    @mock.patch("fedora_messaging.twisted.service.twisted_ssl.Certificate.loadPEM")
    @mock.patch("fedora_messaging.twisted.service.twisted_ssl.optionsForClientTLS")
    def test_no_cert(self, mock_opts, mock_load_pem, fixtures_dir):
        """Assert if there's no client cert, the factory isn't configured to use it."""
        tls_conf = {
            "keyfile": os.path.join(fixtures_dir, "key.pem"),
            "certfile": None,
            "ca_cert": os.path.join(fixtures_dir, "ca_bundle.pem"),
        }

        with mock.patch.dict(config.conf, {"tls": tls_conf}):
            factory = _ssl_context_factory(mock.Mock(host="dummy_host"))

        mock_opts.assert_called_once_with(
            "dummy_host",
            trustRoot=mock_load_pem.return_value,
            clientCertificate=None,
            extraCertificateOptions={"raiseMinimumTo": twisted_ssl.TLSVersion.TLSv1_2},
        )
        assert factory == mock_opts.return_value

    @mock.patch("fedora_messaging.twisted.service.twisted_ssl.PrivateCertificate.loadPEM")
    @mock.patch("fedora_messaging.twisted.service.twisted_ssl.Certificate.loadPEM")
    @mock.patch("fedora_messaging.twisted.service.twisted_ssl.optionsForClientTLS")
    def test_key_and_cert(self, mock_opts, mock_load_pem, mock_priv_load_pem, fixtures_dir):
        """Assert if there's a client key and cert, the factory has both."""
        tls_conf = {
            "keyfile": os.path.join(fixtures_dir, "key.pem"),
            "certfile": os.path.join(fixtures_dir, "cert.pem"),
            "ca_cert": os.path.join(fixtures_dir, "ca_bundle.pem"),
        }

        with mock.patch.dict(config.conf, {"tls": tls_conf}):
            factory = _ssl_context_factory(mock.Mock(host="dummy_host"))

        mock_opts.assert_called_once_with(
            "dummy_host",
            trustRoot=mock_load_pem.return_value,
            clientCertificate=mock_priv_load_pem.return_value,
            extraCertificateOptions={"raiseMinimumTo": twisted_ssl.TLSVersion.TLSv1_2},
        )
        assert factory == mock_opts.return_value



"""The API for publishing messages and consuming from message queues."""

import inspect
import logging

import crochet
from twisted.internet import defer, reactor, threads

from . import config, exceptions
from .message import (
    dumps,
    loads,
    Message,
    SERIALIZED_MESSAGE_SCHEMA,
    SEVERITIES,
)
from .signals import pre_publish_signal, publish_failed_signal, publish_signal
from .twisted import service
from .twisted.consumer import Consumer


_log = logging.getLogger(__name__)


__all__ = (
    "Message",
    "consume",
    "publish",
    "twisted_consume",
    "Consumer",
    "pre_publish_signal",
    "publish_signal",
    "publish_failed_signal",
    "loads",
    "dumps",
    "SERIALIZED_MESSAGE_SCHEMA",
    "SEVERITIES",
)

# The Twisted service that consumers are registered with.
_twisted_service = None


def _init_twisted_service():
    global _twisted_service
    if _twisted_service is None:
        _twisted_service = service.FedoraMessagingServiceV2(config.conf["amqp_url"])
        reactor.callWhenRunning(_twisted_service.startService)
        # Twisted is killing the underlying connection before stopService gets
        # called, so we need to add it as a pre-shutdown event to gracefully
        # finish up messages in progress.
        reactor.addSystemEventTrigger("before", "shutdown", _twisted_service.stopService)


def _check_callback(callback):
    """
    Turns a callback that is potentially a class into a callable object.

    Args:
        callback (object): An object that might be a class, method, or function.
        if the object is a class, this creates an instance of it.

    Raises:
        ValueError: If an instance can't be created or it isn't a callable object.
        TypeError: If the class requires arguments to be instantiated.

    Returns:
        callable: A callable object suitable for use as the consumer callback.
    """
    # If the callback is a class, create an instance of it first
    if inspect.isclass(callback):
        callback_object = callback()
        if not callable(callback_object):
            raise ValueError("Callback must be a class that implements __call__ or a function.")
    elif callable(callback):
        callback_object = callback
    else:
        raise ValueError("Callback must be a class that implements __call__ or a function.")

    return callback_object


def twisted_consume(callback, bindings=None, queues=None):
    """
    Start a consumer using the provided callback and run it using the Twisted
    event loop (reactor).

    .. note:: Callbacks run in a Twisted-managed thread pool using the
        :func:`twisted.internet.threads.deferToThread` API to avoid them blocking
        the event loop. If you wish to use Twisted APIs in your callback you must
        use the :func:`twisted.internet.threads.blockingCallFromThread` or
        :class:`twisted.internet.interfaces.IReactorFromThreads` APIs.

    This API expects the caller to start the reactor.

    Args:
        callback (callable): A callable object that accepts one positional argument,
            a :class:`.Message` or a class object that implements the ``__call__``
            method. The class will be instantiated before use.
        bindings (dict or list of dict): Bindings to declare before consuming. This
            should be the same format as the :ref:`conf-bindings` configuration.
        queues (dict): The queue to declare and consume from. Each key in this
            dictionary should be a queue name to declare, and each value should
            be a dictionary with the "durable", "auto_delete", "exclusive", and
            "arguments" keys.

    Raises:
        ValueError: If the callback, bindings, or queues are invalid.

    Returns:
        twisted.internet.defer.Deferred:
            A deferred that fires with the list of one or more
            :class:`.Consumer` objects. Each consumer object has a
            :attr:`.Consumer.result` instance variable that is a Deferred that
            fires or errors when the consumer halts. Note that this API is
            meant to survive network problems, so consuming will continue until
            :meth:`.Consumer.cancel` is called or a fatal server error occurs.
            The deferred returned by this function may error back with a
            :class:`fedora_messaging.exceptions.BadDeclaration` if queues or
            bindings cannot be declared on the broker, a
            :class:`fedora_messaging.exceptions.PermissionException` if the user
            doesn't have access to the queue, or
            :class:`fedora_messaging.exceptions.ConnectionException` if the TLS
            or AMQP handshake fails.
    """
    if isinstance(bindings, dict):
        bindings = [bindings]
    if bindings is None:
        bindings = config.conf["bindings"]
    else:
        try:
            config.validate_bindings(bindings)
        except exceptions.ConfigurationException as e:
            raise ValueError(e.message) from e

    if queues is None:
        queues = config.conf["queues"]
    else:
        try:
            config.validate_queues(queues)
        except exceptions.ConfigurationException as e:
            raise ValueError(e.message) from e

    callback = _check_callback(callback)

    _init_twisted_service()
    return _twisted_service._service.factory.consume(callback, bindings, queues)


@crochet.run_in_reactor
@defer.inlineCallbacks
def _twisted_consume_wrapper(callback, bindings, queues):
    """
    Wrap the :func:`twisted_consume` function for a synchronous API.

    Returns:
        defer.Deferred: Fires with the consumers once all consumers have halted
            or a consumer encounters an error.
    """
    consumers = yield twisted_consume(callback, bindings=bindings, queues=queues)
    try:
        yield defer.gatherResults([c.result for c in consumers])
    except defer.FirstError as e:
        e.subFailure.raiseException()
    defer.returnValue(consumers)


def consume(callback, bindings=None, queues=None):
    """
    Start a message consumer that executes the provided callback when messages are
    received.

    This API is blocking and will not return until the process receives a signal
    from the operating system.

    .. warning:: This API is runs the callback in the IO loop thread. This means
        if your callback could run for a length of time near the heartbeat interval,
        which is likely on the order of 60 seconds, the broker will kill the TCP
        connection and the message will be re-delivered on start-up.

        For now, use the :func:`twisted_consume` API which runs the
        callback in a thread and continues to handle AMQP events while the
        callback runs if you have a long-running callback.

    The callback receives a single positional argument, the message:

    >>> from fedora_messaging import api
    >>> def my_callback(message):
    ...     print(message)
    >>> bindings = [{'exchange': 'amq.topic', 'queue': 'demo', 'routing_keys': ['#']}]
    >>> queues = {
    ...     "demo": {"durable": False, "auto_delete": True, "exclusive": True, "arguments": {}}
    ... }
    >>> api.consume(my_callback, bindings=bindings, queues=queues)

    If the bindings and queue arguments are not provided, they will be loaded from
    the configuration.

    For complete documentation on writing consumers, see the :ref:`consumers`
    documentation.

    Args:
        callback (callable): A callable object that accepts one positional argument,
            a :class:`Message` or a class object that implements the ``__call__``
            method. The class will be instantiated before use.
        bindings (dict or list of dict): Bindings to declare before consuming. This
            should be the same format as the :ref:`conf-bindings` configuration.
        queues (dict): The queue or queues to declare and consume from. This should be
            in the same format as the :ref:`conf-queues` configuration dictionary where
            each key is a queue name and each value is a dictionary of settings for that
            queue.

    Raises:
        fedora_messaging.exceptions.HaltConsumer: If the consumer requests that
            it be stopped.
        ValueError: If the consumer provides a callback that is not a class that
            implements __call__ and is not a function, if the bindings argument
            is not a dict or list of dicts with the proper keys, or if the queues
            argument isn't a dict with the proper keys.
    """
    crochet.setup()
    eventual_result = _twisted_consume_wrapper(callback, bindings, queues)
    try:
        # Waiting indefinitely on crochet is deprecated, but this is a common value
        # used in the code base which is equivalent to 68 years, which I believe to
        # be an acceptably long amount of time. If you hit this limit you
        # should know that I'll feel no sympathy for you as I'll almost
        # certainly be dead.
        eventual_result.wait(timeout=2**31)
    except (ValueError, exceptions.HaltConsumer):
        raise
    except Exception:
        # https://crochet.readthedocs.io/en/stable/workarounds.html#missing-tracebacks
        _log.error(
            "Consuming raised an unexpected error, please report a bug:\n%s",
            eventual_result.original_failure().getTraceback(),
        )
        raise


@defer.inlineCallbacks
def twisted_publish(message, exchange=None):
    """
    Publish messages via Twisted.

    If you are calling this method from a consumption callback, you should use
    :func:`twisted.internet.threads.blockingCallFromThread`. See:
    https://twisted.org/documents/16.3.0/core/howto/threading.html#invoking-twisted-from-other-threads

    Returns:
        defer.Deferred: A deferred that fires when a message has been published
            and confirmed by the broker.
    """
    if exchange is None:
        exchange = config.conf["publish_exchange"]
    yield threads.deferToThread(pre_publish_signal.send, publish, message=message)
    try:
        yield _twisted_service._service.factory.publish(message, exchange=exchange)
        yield threads.deferToThread(publish_signal.send, publish, message=message)
    except Exception as e:
        yield threads.deferToThread(publish_failed_signal.send, publish, message=message, reason=e)
        raise


@crochet.run_in_reactor
@defer.inlineCallbacks
def _twisted_publish_wrapper(message, exchange):
    """
    Wrapper to provide a synchronous API for publishing messages via Twisted.

    Returns:
        defer.Deferred: A deferred that fires when a message has been published
            and confirmed by the broker.
    """
    _init_twisted_service()
    try:
        yield twisted_publish(message, exchange)
    except defer.CancelledError:
        _log.debug("Canceled publish of %r to %s due to timeout", message, exchange)


def publish(message, exchange=None, timeout=30):
    """
    Publish a message to an exchange.

    This is a synchronous call, meaning that when this function returns, an
    acknowledgment has been received from the message broker and you can be
    certain the message was published successfully.

    There are some cases where an error occurs despite your message being
    successfully published. For example, if a network partition occurs after
    the message is received by the broker. Therefore, you may publish duplicate
    messages. For complete details, see the :ref:`publishing` documentation.

    >>> from fedora_messaging import api
    >>> message = api.Message(body={'Hello': 'world'}, topic='Hi')
    >>> api.publish(message)

    If an attempt to publish fails because the broker rejects the message, it
    is not retried. Connection attempts to the broker can be configured using
    the "connection_attempts" and "retry_delay" options in the broker URL. See
    :class:`pika.connection.URLParameters` for details.

    Args:
        message (message.Message): The message to publish.
        exchange (str): The name of the AMQP exchange to publish to; defaults to
            :ref:`conf-publish-exchange`
        timeout (int): The maximum time in seconds to wait before giving up attempting
            to publish the message. If the timeout is reached, a PublishTimeout exception
            is raised.

    Raises:
        fedora_messaging.exceptions.PublishReturned: Raised if the broker rejects the
            message.
        fedora_messaging.exceptions.PublishTimeout: Raised if the broker could not be
            contacted in the given timeout time.
        fedora_messaging.exceptions.PublishForbidden: Raised if the broker rejects the
            message because of permission issues.
        fedora_messaging.exceptions.ValidationError: Raised if the message
            fails validation with its JSON schema. This only depends on the
            message you are trying to send, the AMQP server is not involved.
    """
    crochet.setup()

    eventual_result = _twisted_publish_wrapper(message, exchange)
    try:
        eventual_result.wait(timeout=timeout)
    except crochet.TimeoutError as e:
        eventual_result.cancel()
        wrapper = exceptions.PublishTimeout(
            f"Publishing timed out after waiting {timeout} seconds."
        )
        raise wrapper from e
    except Exception:
        _log.error(eventual_result.original_failure().getTraceback())
        raise



import os

from twisted.internet import reactor, task


def sleep(delay):
    # Returns a deferred that calls do-nothing function
    # after `delay` seconds
    return task.deferLater(reactor, delay, lambda: None)


RABBITMQ_HOST = os.environ.get("RABBITMQ_HOST", "localhost")



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Tests for the testing utilities."""

from unittest import mock

import pytest

from fedora_messaging import api, message, testing


class CustomMessage(api.Message):
    pass


class TestMockSends:
    """Tests for the :func:`fedora_messaging.testing.mock_sends` function."""

    def test_class(self):
        """Assert all goes well if the message published matches the asserted class."""

        def pub():
            api.publish(api.Message())

        with testing.mock_sends(api.Message) as sent:
            pub()
        assert len(sent) == 1
        assert isinstance(sent[0], api.Message)

    def test_instance(self):
        """Assert all goes well if the message published matches the asserted instance."""

        def pub():
            api.publish(api.Message())

        with testing.mock_sends(api.Message()) as sent:
            pub()
        assert len(sent) == 1
        assert isinstance(sent[0], api.Message)

    def test_expected_none(self):
        """Assert failure if a message is unexpectedly sent."""

        def pub():
            api.publish(api.Message())

        with pytest.raises(AssertionError) as cm:
            with testing.mock_sends() as sent:
                pub()
        assert "Expected 0 messages to be sent, but 1 were sent" == cm.value.args[0]
        assert len(sent) == 1
        assert isinstance(sent[0], api.Message)

    def test_mix_class_instance(self):
        """Assert a mix of class and instance works."""

        def pub():
            api.publish(api.Message())
            api.publish(CustomMessage())

        with mock.patch.dict(message._class_to_schema_name, {CustomMessage: "custom"}):
            with testing.mock_sends(api.Message(), CustomMessage) as sent:
                pub()
        assert len(sent) == 2
        assert isinstance(sent[1], CustomMessage)

    def test_mix_class_instance_order_matters(self):
        """Assert the order of messages matters."""
        expected_err = (
            "Expected message of type <class 'tests.unit.test_testing"
            ".CustomMessage'>, but <class 'fedora_messaging.message.Message'> was sent"
        )

        def pub():
            api.publish(api.Message())
            api.publish(CustomMessage())

        with mock.patch.dict(message._class_to_schema_name, {CustomMessage: "custom"}):
            with pytest.raises(AssertionError) as cm:
                with testing.mock_sends(CustomMessage, api.Message()):
                    pub()
        assert expected_err == cm.value.args[0]

    def test_too_many(self):
        """Assert publishing more messages than expected fails with an AssertionError."""

        def pub():
            api.publish(api.Message())
            api.publish(api.Message())

        with pytest.raises(AssertionError) as cm:
            with testing.mock_sends(api.Message) as sent:
                pub()
        assert "Expected 1 messages to be sent, but 2 were sent" == cm.value.args[0]
        assert len(sent) == 2

    def test_wrong_type(self):
        """Assert sending the wrong type of message raises an AssertionError."""
        expected_err = (
            "Expected message of type <class 'tests.unit.test_testing"
            ".CustomMessage'>, but <class 'fedora_messaging.message.Message'> was sent"
        )

        def pub():
            api.publish(api.Message())

        with pytest.raises(AssertionError) as cm:
            with testing.mock_sends(CustomMessage) as sent:
                pub()
        assert expected_err == cm.value.args[0]
        assert len(sent) == 1
        assert isinstance(sent[0], api.Message)
        assert not isinstance(sent[0], CustomMessage)



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import datetime
import json
from unittest import mock

import jsonschema
import pika
import pytest

from fedora_messaging import exceptions, message


class DeprecatedMessage(message.Message):
    deprecated = True


class TestGetMessage:
    """Tests for the :func:`fedora_messaging.message.get_message` function."""

    def test_missing_severity(self):
        """Assert the default severity is INFO if it's not in the headers."""
        msg = message.Message(severity=message.ERROR)
        del msg._headers["fedora_messaging_severity"]

        recv_msg = message.get_message("", msg._properties, b"{}")
        assert recv_msg.severity == message.INFO

    def test_invalid_severity(self):
        """Assert the invalid severity fails validation."""
        msg = message.Message()
        msg._headers["fedora_messaging_severity"] = 42

        with pytest.raises(exceptions.ValidationError) as excinfo:
            message.get_message("", msg._properties, b"{}")
        assert excinfo.value.summary == "42 is not one of [10, 20, 30, 40]"

    def test_missing_headers(self):
        """Assert missing headers results in a default message."""
        msg = message.Message()
        msg._headers = None

        received_msg = message.get_message(
            msg._encoded_routing_key, msg._properties, msg._encoded_body
        )
        assert isinstance(received_msg, message.Message)

    @mock.patch.dict(message._class_to_schema_name, {DeprecatedMessage: "deprecated_message_id"})
    @mock.patch.dict(message._schema_name_to_class, {"deprecated_message_id": DeprecatedMessage})
    def test_deprecated(self, caplog):
        """Assert a deprecation warning is produced when indicated."""
        msg = DeprecatedMessage(topic="dummy.topic")
        received_msg = message.get_message(msg.topic, msg._properties, msg._encoded_body)
        assert isinstance(received_msg, DeprecatedMessage)
        assert len(caplog.messages) == 1
        assert caplog.messages[0] == (
            "A message with a deprecated schema (tests.unit.test_message.DeprecatedMessage) "
            "has been received on topic 'dummy.topic'. You should check "
            "the emitting application's documentation to upgrade to the newer schema version."
        )


class TestMessageDumps:
    """Tests for the :func:`fedora_messaging.message.dumps` function."""

    def test_proper_message(self):
        """Assert proper json is returned"""
        test_topic = "test topic"
        test_body = {"test_key": "test_value"}
        test_queue = "test queue"
        test_id = "test id"
        test_headers = {
            "fedora_messaging_schema": "base.message",
            "fedora_messaging_severity": message.WARNING,
        }
        test_properties = pika.BasicProperties(
            content_type="application/json",
            content_encoding="utf-8",
            delivery_mode=2,
            headers=test_headers,
            message_id=test_id,
            priority=2,
        )
        test_msg = message.Message(body=test_body, topic=test_topic, properties=test_properties)

        test_msg.queue = test_queue
        expected_json = (
            '{"body": {"test_key": "test_value"}, "headers": {"fedora_messaging_schema": '
            '"base.message", "fedora_messaging_severity": 30}, "id": "test id", '
            '"priority": 2, "queue": "test queue", "topic": "test topic"}\n'
        )
        assert expected_json == message.dumps(test_msg)

    def test_proper_message_multiple(self):
        """Assert proper json is returned"""
        test_topic = "test topic"
        test_body = {"test_key": "test_value"}
        test_queue = "test queue"
        test_id = "test id"
        test_headers = {
            "fedora_messaging_schema": "base.message",
            "fedora_messaging_severity": message.WARNING,
        }
        test_properties = pika.BasicProperties(
            content_type="application/json",
            content_encoding="utf-8",
            delivery_mode=2,
            headers=test_headers,
            message_id=test_id,
        )
        test_msg = message.Message(body=test_body, topic=test_topic, properties=test_properties)
        test_msg2 = message.Message(body=test_body, topic=test_topic, properties=test_properties)
        test_msg.queue = test_queue
        test_msg2.queue = test_queue
        expected_json = (
            '{"body": {"test_key": "test_value"}, "headers": {"fedora_messaging_schema": '
            '"base.message", "fedora_messaging_severity": 30}, "id": "test id", '
            '"priority": 0, "queue": "test queue", "topic": "test topic"}\n'
            '{"body": {"test_key": "test_value"}, "headers": {"fedora_messaging_schema": '
            '"base.message", "fedora_messaging_severity": 30}, "id": "test id", '
            '"priority": 0, "queue": "test queue", "topic": "test topic"}\n'
        )

        assert expected_json == message.dumps([test_msg, test_msg2])

    def test_improper_messages(self):
        """Assert TypeError is raised when improper messages are provided"""
        messages = ["m1", "m2"]
        with pytest.raises(exceptions.ValidationError) as excinfo:
            message.dumps(messages)
        assert excinfo.value.summary == "'str' object has no attribute 'validate'"


class TestMessageLoads:
    """Tests for the :func:`fedora_messaging.message.loads` function."""

    def test_proper_json(self):
        """Assert loading single message from json work."""
        message_json = (
            '{"topic": "test topic", "headers": {"fedora_messaging_schema": "base.message", '
            '"fedora_messaging_severity": 30}, "id": "test id", "body": '
            '{"test_key": "test_value"}, "priority": 2, "queue": "test queue"}\n'
        )
        messages = message.loads(message_json)
        assert len(messages) == 1
        test_message = messages[0]
        assert isinstance(test_message, message.Message)
        assert "test topic" == test_message.topic
        assert "test id" == test_message.id
        assert {"test_key": "test_value"} == test_message.body
        assert "test queue" == test_message.queue
        assert 2 == test_message.priority
        assert message.WARNING == test_message._headers["fedora_messaging_severity"]
        assert "base.message" == test_message._headers["fedora_messaging_schema"]

    def test_improper_json(self):
        """Assert proper exception is raised when improper json is provided."""
        message_json = "improper json"
        with pytest.raises(exceptions.ValidationError):
            message.loads(message_json)

    def test_missing_headers(self):
        """Assert no exception is raised when headers are missing."""
        message_dict = {
            "topic": "test topic",
            "id": "test id",
            "body": {"test_key": "test_value"},
            "queue": "test queue",
        }
        test_message = message.load_message(message_dict)
        assert test_message._headers["fedora_messaging_schema"] == "base.message"
        assert test_message._headers["fedora_messaging_severity"] == message.INFO
        assert "sent-at" in test_message._headers

    def test_missing_messaging_schema(self):
        """Assert the default schema is used when messaging schema is missing."""
        message_dict = {
            "id": "test id",
            "topic": "test topic",
            "headers": {"fedora_messaging_severity": 30},
            "body": {"test_key": "test_value"},
            "queue": "test queue",
        }
        test_message = message.load_message(message_dict)
        assert isinstance(test_message, message.Message)

    def test_missing_body(self):
        """Assert proper exception is raised when body is missing."""
        message_dict = {
            "id": "test id",
            "topic": "test topic",
            "headers": {
                "fedora_messaging_schema": "base.message",
                "fedora_messaging_severity": 30,
            },
            "queue": "test queue",
        }
        with pytest.raises(exceptions.ValidationError):
            message.load_message(message_dict)

    @mock.patch("fedora_messaging.message.uuid")
    def test_missing_id(self, uuid):
        """Assert proper exception is raised when id is missing."""
        uuid.uuid4.return_value = "dummy-uuid"
        message_dict = {
            "topic": "test topic",
            "headers": {
                "fedora_messaging_schema": "base.message",
                "fedora_messaging_severity": 30,
            },
            "body": {"test_key": "test_value"},
            "queue": "test queue",
        }
        test_message = message.load_message(message_dict)
        assert test_message.id == "dummy-uuid"

    def test_missing_queue(self):
        """Assert message without queue is accepted and the queue is set to None."""
        message_dict = {
            "id": "test id",
            "topic": "test topic",
            "headers": {
                "fedora_messaging_schema": "base.message",
                "fedora_messaging_severity": 30,
            },
            "body": {"test_key": "test_value"},
        }
        test_message = message.load_message(message_dict)
        assert test_message.queue is None

    def test_missing_topic(self):
        """Assert proper exception is raised when topic is missing."""
        message_dict = {
            "id": "test id",
            "headers": {
                "fedora_messaging_schema": "base.message",
                "fedora_messaging_severity": 30,
            },
            "body": {"test_key": "test_value"},
            "queue": "test queue",
        }
        with pytest.raises(exceptions.ValidationError):
            message.load_message(message_dict)

    def test_missing_severity(self):
        """Assert no exception is raised when severity is missing."""
        message_dict = {
            "topic": "test topic",
            "headers": {"fedora_messaging_schema": "base.message"},
            "id": "test id",
            "body": {"test_key": "test_value"},
            "queue": "test queue",
        }
        test_message = message.load_message(message_dict)
        assert test_message.severity == message.INFO

    def test_missing_priority(self):
        """Assert message without priority is accepted and the priority is set to zero."""
        message_dict = {
            "id": "test id",
            "topic": "test topic",
            "headers": {
                "fedora_messaging_schema": "base.message",
                "fedora_messaging_severity": 30,
            },
            "body": {"test_key": "test_value"},
            "queue": "test queue",
        }
        test_message = message.load_message(message_dict)
        assert test_message.priority == 0

    def test_validation_failure(self):
        """Assert proper exception is raised when message validation failed."""
        message_dict = {
            "id": "test id",
            "topic": "test topic",
            "headers": {
                "fedora_messaging_schema": "base.message",
                "fedora_messaging_severity": 41,
            },
            "body": {"test_key": "test_value"},
            "queue": "test queue",
        }
        with pytest.raises(exceptions.ValidationError):
            message.load_message(message_dict)


class TestMessage:
    """Tests for the :class:`fedora_messaging.message.Message` class."""

    def test_summary(self):
        """Assert message summaries default to the message topic."""
        msg = message.Message(topic="test.topic")
        assert msg.topic == msg.summary

    def test_str(self):
        """Assert calling str on a message produces a human-readable result."""
        msg = message.Message(topic="test.topic", body={"my": "key"})
        expected_headers = json.dumps(
            msg._headers, sort_keys=True, indent=4, separators=(",", ": ")
        )
        expected = (
            f"Id: {msg.id}\nTopic: test.topic\n"
            f"Headers: {expected_headers}"
            '\nBody: {\n    "my": "key"\n}'
        )
        assert expected == str(msg)

    def test_equality(self):
        """
        Assert two messages of the same class with the same topic, headers, and
        body are equivalent.
        """
        assert message.Message(topic="test.topic", body={"my": "key"}) == message.Message(
            topic="test.topic", body={"my": "key"}
        )

    def test_equality_different_sent_at(self):
        """Assert the "sent-at" key is not included in the equality check."""
        m1 = message.Message(topic="test.topic", body={"my": "key"})
        m2 = message.Message(topic="test.topic", body={"my": "key"})
        m2._headers["sent-at"] = datetime.datetime(1970, 1, 2).isoformat()

        assert m1 == m2

    def test_repr(self):
        """Assert the message produces a valid representation of the message."""
        msg = message.Message(topic="test.topic", body={"my": "key"})
        expected = f"Message(id='{msg.id}', topic='test.topic', body={{'my': 'key'}})"
        assert expected == repr(msg)

    def test_valid_message(self):
        """Assert that the default schema allows objects for the header and body."""
        message.Message(topic="test.topic", headers={}, body={}).validate()

    def test_invalid_message(self):
        """Assert that a non-object body raises a ValidationError on validation."""
        msg = message.Message(topic="test.topic", headers={}, body="text")
        with pytest.raises(jsonschema.ValidationError):
            msg.validate()

    def test_default_message(self):
        """Assert the default message is valid."""
        message.Message().validate()

    def test_properties_default(self):
        msg = message.Message()
        assert msg._properties.content_type == "application/json"
        assert msg._properties.content_encoding == "utf-8"
        assert msg._properties.delivery_mode == 2
        assert "sent-at" in msg._properties.headers
        assert "fedora_messaging_schema" in msg._properties.headers
        assert msg._properties.headers["fedora_messaging_schema"] == "base.message"

    def test_headers(self):
        msg = message.Message(headers={"foo": "bar"})
        assert "foo" in msg._properties.headers
        assert msg._properties.headers["foo"] == "bar"
        # The fedora_messaging_schema key must also be added when headers are given.
        assert msg._properties.headers["fedora_messaging_schema"] == "base.message"

    def test_severity_default_header_set(self):
        """Assert the default severity is placed in the header if unspecified."""
        assert message.Message.severity == message.INFO
        msg = message.Message()
        assert msg._headers["fedora_messaging_severity"] == message.INFO

    def test_severity_custom_header_set(self):
        """Assert custom severity setting is placed in the header."""
        assert message.Message.severity == message.INFO
        msg = message.Message(severity=message.ERROR)
        assert msg._headers["fedora_messaging_severity"] == message.ERROR

    def test_sent_at(self):
        """Assert a timestamp is inserted and contains explicit timezone information."""
        mock_datetime = mock.Mock()
        mock_datetime.now.return_value = datetime.datetime(
            1970, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc
        )

        with mock.patch("datetime.datetime", mock_datetime):
            msg = message.Message()

        assert "1970-01-01T00:00:00+00:00" == msg._headers["sent-at"]

    def test_priority(self):
        """Assert is set correctly."""
        msg = message.Message()
        assert 0 == msg.priority
        assert 0 == msg._headers["priority"]
        msg.priority = 42
        assert 42 == msg.priority
        assert 42 == msg._headers["priority"]
        msg.priority = None
        assert 0 == msg.priority
        assert 0 == msg._properties.priority
        assert 0 == msg._headers["priority"]

    def test_properties(self):
        properties = object()
        msg = message.Message(properties=properties)
        assert msg._properties == properties

    def test_encoded_routing_key(self):
        """Assert encoded routing key is correct."""
        msg = message.Message(topic="test.topic")
        assert msg._encoded_routing_key == b"test.topic"

    def test_encoded_body(self):
        """Assert encoded body is correct."""
        body = {"foo": "barr\u00e9"}
        msg = message.Message(body=body)
        assert msg._encoded_body == json.dumps(body).encode("utf-8")

    def test_url(self):
        # The url property must exist and defaults to None
        assert message.Message().url is None

    def test_app_name(self):
        # The app_name property must exist and defaults to None
        assert message.Message().app_name is None

    def test_app_icon(self):
        # The app_icon property must exist and defaults to None
        assert message.Message().app_icon is None

    def test_agent_name(self):
        # The agent_name property must exist and defaults to None
        assert message.Message().agent_name is None

    def test_agent_avatar(self):
        # The agent_avatar property must exist and defaults to None
        assert message.Message().agent_avatar is None

    def test_usernames(self):
        # The usenames property must exist and be a list
        assert message.Message().usernames == []

    def test_groups(self):
        # The groups property must exist and be a list
        assert message.Message().groups == []

    def test_packages(self):
        # The packages property must exist and be a list
        assert message.Message().packages == []

    def test_containers(self):
        """The containers attribute must exist and be a list."""
        assert message.Message().containers == []

    def test_modules(self):
        """The modules attribute must exist and be a list."""
        assert message.Message().modules == []

    def test_flatpaks(self):
        """The flatpaks attribute must exist and be a list."""
        assert message.Message().flatpaks == []

    def test_topic_prefix(self):
        """Assert the topic prefix is used in the encoded routing key."""
        with mock.patch.dict(message.config.conf, {"topic_prefix": "prefix"}):
            msg = message.Message(topic="test.topic")
            assert msg._encoded_routing_key == b"prefix.test.topic"


class CustomMessage(message.Message):
    """Test class that returns values for filter properties."""

    @property
    def usernames(self):
        try:
            return self.body["users"]
        except KeyError:
            return []

    @property
    def groups(self):
        try:
            return self.body["groups"]
        except KeyError:
            return []

    @property
    def packages(self):
        try:
            return self.body["packages"]
        except KeyError:
            return []

    @property
    def containers(self):
        try:
            return self.body["containers"]
        except KeyError:
            return []
        pass

    @property
    def modules(self):
        try:
            return self.body["modules"]
        except KeyError:
            return []

    @property
    def flatpaks(self):
        try:
            return self.body["flatpaks"]
        except KeyError:
            return []


@mock.patch.dict(message._class_to_schema_name, {CustomMessage: "custom_id"})
class TestCustomMessage:
    """Tests for a Message subclass that provides filter headers"""

    def test_usernames(self):
        """Assert usernames are placed in the message headers."""
        msg = CustomMessage(body={"users": ["jcline", "abompard"]})

        assert msg.usernames == ["jcline", "abompard"]
        assert "fedora_messaging_user_jcline" in msg._headers
        assert "fedora_messaging_user_abompard" in msg._headers

    def test_groups(self):
        """Assert groups are placed in the message headers."""
        msg = CustomMessage(body={"groups": ["fedora-infra", "copr"]})

        assert msg.groups == ["fedora-infra", "copr"]
        assert "fedora_messaging_group_fedora-infra" in msg._headers
        assert "fedora_messaging_group_copr" in msg._headers

    def test_packages(self):
        """Assert RPM packages are placed in the message headers."""
        msg = CustomMessage(body={"packages": ["kernel", "python-requests"]})

        assert msg.packages == ["kernel", "python-requests"]
        assert "fedora_messaging_rpm_kernel" in msg._headers
        assert "fedora_messaging_rpm_python-requests" in msg._headers

    def test_containers(self):
        """Assert containers are placed in the message headers."""
        msg = CustomMessage(body={"containers": ["rawhide:latest", "f29"]})

        assert msg.containers == ["rawhide:latest", "f29"]
        assert "fedora_messaging_container_rawhide:latest" in msg._headers
        assert "fedora_messaging_container_f29" in msg._headers

    def test_modules(self):
        """Assert modules are placed in the message headers."""
        msg = CustomMessage(body={"modules": ["nodejs", "ripgrep"]})

        assert msg.modules == ["nodejs", "ripgrep"]
        assert "fedora_messaging_module_nodejs" in msg._headers
        assert "fedora_messaging_module_ripgrep" in msg._headers

    def test_flatpaks(self):
        """Assert flatpaks are placed in the message headers."""
        msg = CustomMessage(body={"flatpaks": ["firefox", "hexchat"]})

        assert msg.flatpaks == ["firefox", "hexchat"]
        assert "fedora_messaging_flatpak_firefox" in msg._headers
        assert "fedora_messaging_flatpak_hexchat" in msg._headers


class CustomValidatedMessage(message.Message):
    """Test class where filter properties depend on the validation."""

    body_schema = {
        "id": "http://fedoraproject.org/message-schema/custom-validated#",
        "$schema": "http://json-schema.org/draft-04/schema#",
        "type": "object",
        "properties": {
            "users": {
                "type": "array",
                "items": {
                    "type": "string",
                },
            },
        },
        "required": ["user"],
    }

    @property
    def usernames(self):
        return self.body["users"]


@mock.patch.dict(message._class_to_schema_name, {CustomValidatedMessage: "custom_validated_id"})
class TestCustomValidatedMessage:
    """Tests for CustomValidatedMessage"""

    def test_usernames(self):
        """Assert the instantiation works even if the message is invalid."""
        try:
            msg = CustomValidatedMessage(body={})
        except KeyError:
            pytest.fail("Error in filter properties prevented the message from being instanciated.")
        with pytest.raises(jsonschema.ValidationError):
            msg.validate()


class TestClassRegistry:
    """Tests for the :func:`fedora_messaging.message.load_message_classes`."""

    def test_load_message_name_to_class(self):
        """Assert the entry point name maps to the class object."""
        with mock.patch.dict(message._schema_name_to_class, {}, clear=True):
            message.load_message_classes()
            assert "base.message" in message._schema_name_to_class
            assert message._schema_name_to_class["base.message"] is message.Message

    def test_load_message_class_to_name(self):
        """Assert the entry point name maps to the class object."""
        with mock.patch.dict(message._class_to_schema_name, {}, clear=True):
            message.load_message_classes()
            assert message.Message in message._class_to_schema_name
            assert "base.message" == message._class_to_schema_name[message.Message]

    @mock.patch("fedora_messaging.message._registry_loaded", False)
    def test_get_class_autoload(self):
        """Assert the registry is automatically loaded."""
        with mock.patch.dict(message._schema_name_to_class, {}, clear=True):
            assert message.get_class("base.message") == message.Message

    @mock.patch("fedora_messaging.message._registry_loaded", True)
    def test_get_class_default(self):
        """Assert the base class is returns if the class is unknown."""
        with mock.patch.dict(message._schema_name_to_class, {}, clear=True):
            assert message.get_class("no.such.message") == message.Message

    @mock.patch("fedora_messaging.message._registry_loaded", False)
    def test_get_name_autoload(self):
        """Assert the registry is automatically loaded."""
        with mock.patch.dict(message._class_to_schema_name, {}, clear=True):
            assert message.get_name(message.Message) == "base.message"

    @mock.patch("fedora_messaging.message._registry_loaded", True)
    def test_get_name_autoload_once(self):
        """Assert the registry doesn't repeatedly load."""
        with mock.patch.dict(message._class_to_schema_name, {}, clear=True):
            with pytest.raises(TypeError):
                message.get_name("this.is.not.an.entrypoint")



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
A Twisted Factory for creating and configuring instances of the
:class:`.FedoraMessagingProtocolV2`.

A factory is used to implement automatic re-connections by producing protocol
instances (connections) on demand. Twisted uses factories for its services APIs.

See the `Twisted client
<https://twistedmatrix.com/documents/current/core/howto/clients.html#clientfactory>`_
documentation for more information.
"""


import collections
import logging
from collections import namedtuple

import pika
from twisted.internet import defer, error, protocol
from twisted.python.failure import Failure

from ..exceptions import ConnectionException
from .protocol import FedoraMessagingProtocolV2


_std_log = logging.getLogger(__name__)


def _remap_queue_name(bindings, queue_name):
    """Reset the queue name in a binding, because the queue name might be server-generated.

    Args:
        bindings (dict): A binding description with a ``queue`` attribute.
        queue_name (str): The name of the queue to use.
    """
    for binding in bindings:
        binding["queue"] = queue_name
    # The dicts are changed in-place, don't return anything to make that clear.


#: A namedtuple representing the record of an existing AMQP consumer with its config items.
#:
#: * The ``consumer`` field is the consumer object
#:   (:class:`fedora_messaging.twisted.consumer.Consumer`).
#: * The ``queue`` field is the configuration dict for the queue it's consuming from.
#: * The ``bindings`` field is the list of configuration dicts for the queue's bindings.
ConsumerRecord = namedtuple("ConsumerRecord", ["consumer", "queue", "bindings"])


class FedoraMessagingFactoryV2(protocol.ReconnectingClientFactory):
    """Reconnecting factory for the Fedora Messaging protocol."""

    def __init__(self, parameters, confirms=True):
        """
        Create a new factory for protocol objects.

        Any exchanges, queues, or bindings provided here will be declared and
        set up each time a new protocol instance is created. In other words,
        each time a new connection is set up to the broker, it will start with
        the declaration of these objects.

        Args:
            parameters (pika.ConnectionParameters): The connection parameters.
            confirms (bool): If true, attempt to turn on publish confirms extension.
        """
        self.confirms = confirms
        self.protocol = FedoraMessagingProtocolV2
        self._parameters = parameters
        # Used to implement the when_connected API
        self._client_deferred = defer.Deferred()
        self._client = None
        self._consumers = []

    def __repr__(self):
        """Return the representation of the factory as a string"""
        return f"FedoraMessagingFactoryV2(parameters={self._parameters}, confirms={self.confirms})"

    def buildProtocol(self, addr):
        """Create the Protocol instance.

        See the documentation of
        `twisted.internet.protocol.ReconnectingClientFactory` for details.
        """
        client = self.protocol(self._parameters, confirms=self.confirms)
        client.factory = self

        @defer.inlineCallbacks
        def on_ready(unused_param=None):
            """Reset the connection delay when the AMQP handshake is complete."""
            _std_log.debug("AMQP handshake completed; connection ready for use")
            self.resetDelay()
            self._client = client
            self._client_deferred.callback(client)
            # Renew the deferred to handle reconnections.
            self._client_deferred = defer.Deferred()

            # Restart any consumer from previous connections that wasn't canceled
            # including queues and bindings, as the queue might not have been durable
            for record in self._consumers:
                _std_log.info("Re-registering the %r consumer", record.consumer)
                queue_name = yield client.declare_queue(record.queue)
                _remap_queue_name(record.bindings, queue_name)
                yield client.bind_queues(record.bindings)
                yield client.consume(record.consumer.callback, queue_name, record.consumer)

        def on_ready_connection_errback(failure):
            """If opening the connection fails or is lost, this errback is called."""
            r = failure.trap(
                pika.exceptions.AMQPConnectionError,
                error.ConnectionDone,
                error.ConnectionLost,
            )

            if r == error.ConnectionLost:
                msg = (
                    "The network connection to the broker was lost in a non-clean fashion (%r);"
                    " the connection should be restarted by Twisted."
                )
            else:
                # In this case the connection failed to open. This will be called
                # if the TLS handshake goes wrong (likely) and it may be called if
                # the AMQP handshake fails. It's *probably* a problem with the
                # credentials.
                msg = (
                    "The TCP connection appears to have started, but the TLS or AMQP handshake "
                    "with the broker failed; check your connection and authentication "
                    "parameters and ensure your user has permission to access the vhost"
                )

            wrapped_failure = Failure(
                exc_value=ConnectionException(reason=msg, original=failure),
                exc_type=ConnectionException,
            )
            self._client_deferred.errback(wrapped_failure)
            # Renew the deferred to handle reconnections.
            self._client_deferred = defer.Deferred()

        def general_errback(failure):
            _std_log.error(
                "The connection failed with an unexpected exception; please report this bug: %s",
                failure.getTraceback(),
            )
            self._client_deferred.errback(failure)
            # Renew the deferred to handle reconnections.
            self._client_deferred = defer.Deferred()

        client.ready.addCallback(on_ready)
        client.ready.addErrback(on_ready_connection_errback)
        client.ready.addErrback(general_errback)
        return client

    @defer.inlineCallbacks
    def stopFactory(self):
        """Stop the factory.

        See the documentation of
        `twisted.internet.protocol.ReconnectingClientFactory` for details.
        """
        if self._client:
            yield self._client.halt()
        protocol.ReconnectingClientFactory.stopFactory(self)

    @defer.inlineCallbacks
    def when_connected(self):
        """
        Retrieve the currently-connected Protocol, or the next one to connect.

        Returns:
            defer.Deferred: A Deferred that fires with a connected
                :class:`FedoraMessagingProtocolV2` instance. This is similar to
                the whenConnected method from the Twisted endpoints APIs, which
                is sadly isn't available before 16.1.0, which isn't available
                in EL7.
        """
        if self._client and not self._client.is_closed:
            _std_log.debug("Already connected with %r", self._client)
        else:
            self._client = None
            _std_log.debug("Waiting for %r to fire with new connection", self._client_deferred)
            try:
                yield self._client_deferred
            except defer.CancelledError:
                # Renew the deferred to handle future connections.
                self._client_deferred = defer.Deferred()
                raise
        defer.returnValue(self._client)

    @defer.inlineCallbacks
    def publish(self, message, exchange):
        """
        Publish a :class:`fedora_messaging.message.Message` to an `exchange`_
        on the message broker. This call will survive connection failures and try
        until it succeeds or is canceled.

        Args:
            message (message.Message): The message to publish.
            exchange (str): The name of the AMQP exchange to publish to.

        returns:
            defer.Deferred: A deferred that fires when the message is published.

        Raises:
            NoFreeChannels: If there are no available channels on this connection.
                If this occurs, you can either reduce the number of consumers on this
                connection or create an additional connection.
            PublishReturned: If the published message is rejected by the broker.
            PublishForbidden: If the published message is rejected by the broker because
                of permission issues.
            ConnectionException: If a connection error occurs while publishing. Calling
                this method again will wait for the next connection and publish when it
                is available.

        .. _exchange: https://www.rabbitmq.com/tutorials/amqp-concepts.html#exchanges
        """
        while True:
            protocol = yield self.when_connected()
            try:
                yield protocol.publish(message, exchange)
                break
            except ConnectionException:
                _std_log.info("Publish failed on %r, waiting for new connection", protocol)

    @defer.inlineCallbacks
    def consume(self, callback, bindings, queues):
        """
        Start a consumer that lasts across individual connections.

        Args:
            callback (callable): A callable object that accepts one positional argument,
                a :class:`Message` or a class object that implements the ``__call__``
                method. The class will be instantiated before use.
            bindings (dict or list of dict): Bindings to declare before consuming. This
                should be the same format as the :ref:`conf-bindings` configuration.
            queues (dict): The queues to declare and consume from. Each key in this
                dictionary is a queue, and each value is its settings as a dictionary.
                These settings dictionaries should have the "durable", "auto_delete",
                "exclusive", and "arguments" keys. Refer to :ref:`conf-queues` for
                details on their meanings.

        Returns:
            defer.Deferred:
                A deferred that fires with the list of one or more
                :class:`fedora_messaging.twisted.consumer.Consumer` objects.
                These can be passed to the
                :meth:`FedoraMessagingFactoryV2.cancel` API to halt them. Each
                consumer object has a ``result`` instance variable that is a
                Deferred that fires or errors when the consumer halts. The
                Deferred may error back with a BadDeclaration if the user does
                not have permissions to consume from the queue.
        """
        expanded_bindings = collections.defaultdict(list)
        for binding in bindings:
            for key in binding["routing_keys"]:
                b = binding.copy()
                del b["routing_keys"]
                b["routing_key"] = key
                if "queue" not in b:
                    b["queue"] = ""
                expanded_bindings[b["queue"]].append(b)

        expanded_queues = []
        for name, settings in queues.items():
            q = {"queue": name}
            q.update(settings)
            expanded_queues.append(q)

        protocol = yield self.when_connected()

        consumers = []
        for queue in expanded_queues:
            config_queue_name = queue["queue"]
            queue_name = yield protocol.declare_queue(queue)
            ebs = expanded_bindings.get(config_queue_name, [])
            _remap_queue_name(ebs, queue_name)
            yield protocol.bind_queues(ebs)
            consumer = yield protocol.consume(callback, queue_name)
            self._consumers.append(ConsumerRecord(consumer=consumer, queue=queue, bindings=ebs))
            consumers.append(consumer)

        defer.returnValue(consumers)

    @defer.inlineCallbacks
    def cancel(self, consumers):
        """
        Cancel a consumer that was previously started with consume.

        Args:
            consumer (list of fedora_messaging.api.Consumer): The consumers to cancel.
        """
        for consumer in consumers:
            self._forget_consumer(consumer.queue)
            protocol = yield self.when_connected()
            yield protocol.cancel(consumer)

    def _forget_consumer(self, queue):
        """Forget about a consumer.

        Args:
            queue (str): Forget the consumers that consume from this queue.
        """
        self._consumers = [record for record in self._consumers if record.consumer.queue != queue]



# Copyright (C) 2018  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""
The ``schema_utils`` module contains utilities that may be useful when writing
the Python API of your message schemas.
"""

import collections
from hashlib import sha256
from urllib import parse


def user_avatar_url(username, size=64, default="retro"):
    """Get the avatar URL of the provided Fedora username.

    The URL is returned from the Libravatar service.

    Args:
        username (str): The username to get the avatar of.
        size (int): Size of the avatar in pixels (it's a square).
        default (str): Default avatar to return if not found.
    Returns:
        str: The URL to the avatar image.
    """
    openid = f"http://{username}.id.fedoraproject.org/"
    return libravatar_url(openid=openid, size=size, default=default)


def libravatar_url(email=None, openid=None, size=64, default="retro"):
    """Get the URL to an avatar from libravatar.

    Either the user's email or openid must be provided.

    If you want to use Libravatar federation (through DNS), you should install
    and use the ``libravatar`` library instead. Check out the
    ``libravatar.libravatar_url()`` function.

    Args:
        email (str): The user's email
        openid (str): The user's OpenID
        size (int): Size of the avatar in pixels (it's a square).
        default (str): Default avatar to return if not found.
    Returns:
        str: The URL to the avatar image.
    Raises:
        ValueError: If neither email nor openid are provided.
    """
    # We use an OrderedDict here to make testing easier (URL strings become
    # predictable).
    params = collections.OrderedDict([("s", size), ("d", default)])
    query = parse.urlencode(params)
    if email:
        value = email
    elif openid:
        value = openid
    else:
        raise ValueError("You must provide either the email or the openid.")
    idhash = sha256(value.encode("utf-8")).hexdigest()
    return f"https://seccdn.libravatar.org/avatar/{idhash}?{query}"



"""
This module defines the base class of message objects and keeps a registry of
known message implementations. This registry is populated from Python entry
points in the "fedora.messages" group.

To implement your own message schema, simply create a class that inherits the
:class:`Message` class, and add an entry point in your Python package under the
"fedora.messages" group. For example, an entry point for the :class:`Message`
schema would be::

    entry_points = {
        'fedora.messages': [
            'base.message=fedora_messaging.message:Message'
        ]
    }

The entry point name must be unique to your application and is used to map
messages to your message class, so it's best to prefix it with your application
name (e.g. ``bodhi.new_update_messageV1``). When publishing, the Fedora
Messaging library will add a header with the entry point name of the class used
so the consumer can locate the correct schema.

Since every client needs to have the message schema installed, you should define
this class in a small Python package of its own.
"""

import datetime
import json
import logging
import uuid
from importlib.metadata import entry_points

import jsonschema
import pika

from . import config
from .exceptions import ValidationError
from .schema_utils import user_avatar_url


#: Indicates the message is for debugging or is otherwise very low priority. Users
#: will not be notified unless they've explicitly requested DEBUG level messages.
DEBUG = 10

#: Indicates the message is informational. End users will not receive notifications
#: for these messages by default. For example, automated tests passed for their
#: package.
INFO = 20

#: Indicates a problem or an otherwise important problem. Users are notified of
#: these messages when they pertain to packages they are associated with by default.
#: For example, one or more automated tests failed against their package.
WARNING = 30

#: Indicates a critically important message that users should act upon as soon as
#: possible. For example, their package no longer builds.
ERROR = 40

#: A tuple of all valid severity levels
SEVERITIES = (DEBUG, INFO, WARNING, ERROR)


_log = logging.getLogger(__name__)

# Maps string names of message types to classes and back
_schema_name_to_class = {}
_class_to_schema_name = {}

# Used to load the registry automatically on first use
_registry_loaded = False


def get_class(schema_name):
    """
    Retrieve the message class associated with the schema name.

    If no match is found, the default schema is returned and a warning is logged.

    Args:
        schema_name (str): The name of the :class:`Message` sub-class;
            this is typically the Python path.

    Returns:
        Message: A sub-class of :class:`Message` to create the message from.
    """
    global _registry_loaded
    if not _registry_loaded:
        load_message_classes()

    try:
        return _schema_name_to_class[schema_name]
    except KeyError:
        _log.warning(
            'The schema "%s" is not in the schema registry! Either install '
            "the package with its schema definition or define a schema. "
            "Falling back to the default schema...",
            schema_name,
        )
        return Message


def get_name(cls):
    """
    Retrieve the schema name associated with a message class.

    Returns:
        str: The schema name.

    Raises:
        TypeError: If the message class isn't registered. Check your entry point
            for correctness.
    """
    global _registry_loaded
    if not _registry_loaded:
        load_message_classes()

    try:
        return _class_to_schema_name[cls]
    except KeyError as e:
        raise TypeError(
            f"The class {cls!r} is not in the message registry, which indicates it is"
            ' not in the current list of entry points for "fedora_messaging".'
            " Please check that the class has been added to your package's entry points."
        ) from e


def load_message_classes():
    """Load the 'fedora.messages' entry points and register the message classes."""
    try:
        eps = entry_points(group="fedora.messages")
    except TypeError:
        # Python < 3.10
        # https://docs.python.org/3.10/library/importlib.metadata.html#entry-points
        eps = entry_points().get("fedora.messages", [])
    for message in eps:
        cls = message.load()
        _log.debug(
            "Registering the '%s' key as the '%r' class in the Message " "class registry",
            message.name,
            cls,
        )
        _schema_name_to_class[message.name] = cls
        _class_to_schema_name[cls] = message.name
    global _registry_loaded
    _registry_loaded = True


def get_message(routing_key, properties, body):
    """
    Construct a Message instance given the routing key, the properties and the
    body received from the AMQP broker.

    Args:
        routing_key (str): The AMQP routing key (will become the message topic)
        properties (pika.BasicProperties): the AMQP properties
        body (bytes): The encoded message body

    Raises:
        ValidationError: If Message validation failed or message body
            docoding/loading is impossible.
    """
    if properties.headers is None:
        _log.error(
            "Message (body=%r) arrived without headers. " "A publisher is misbehaving!",
            body,
        )
        properties.headers = {}

    try:
        MessageClass = get_class(properties.headers["fedora_messaging_schema"])
    except KeyError:
        _log.error(
            "Message (headers=%r, body=%r) arrived without a schema header."
            " A publisher is misbehaving!",
            properties.headers,
            body,
        )
        MessageClass = Message

    try:
        severity = properties.headers["fedora_messaging_severity"]
    except KeyError:
        _log.error(
            "Message (headers=%r, body=%r) arrived without a severity."
            " A publisher is misbehaving! Defaulting to INFO.",
            properties.headers,
            body,
        )
        severity = INFO

    if properties.content_encoding is None:
        _log.error("Message arrived without a content encoding")
        properties.content_encoding = "utf-8"
    try:
        body = body.decode(properties.content_encoding)
    except UnicodeDecodeError as e:
        _log.error(
            "Unable to decode message body %r with %s content encoding",
            body,
            properties.content_encoding,
        )
        raise ValidationError(e) from e

    try:
        body = json.loads(body)
    except ValueError as e:
        _log.error("Failed to load message body %r, %r", body, e)
        raise ValidationError(e) from e

    message = MessageClass(body=body, topic=routing_key, properties=properties, severity=severity)
    try:
        message.validate()
        _log.debug("Successfully validated message %r", message)
    except jsonschema.exceptions.ValidationError as e:
        _log.error("Message validation of %r failed: %r", message, e)
        raise ValidationError(e) from e

    if MessageClass.deprecated:
        _log.warning(
            "A message with a deprecated schema (%s.%s) has been received on topic %r. "
            "You should check the emitting application's documentation to upgrade to "
            "the newer schema version.",
            MessageClass.__module__,
            MessageClass.__name__,
            message.topic,
        )
    return message


class Message:
    """
    Messages are simply JSON-encoded objects. This allows message authors to
    define a schema and implement Python methods to abstract the raw message
    from the user. This allows the schema to change and evolve without breaking
    the user-facing API.

    There are a number of properties that are intended to be overridden by
    users.  These fields are used to sort messages for notifications or are
    used to create human-readable versions of the messages. Properties that are
    intended for this purpose are noted in their attribute documentation below.

    Args:
        headers (dict): A set of message headers. Consult the headers schema for
            expected keys and values.
        body (dict): The message body. Consult the body schema for expected keys
            and values. This dictionary must be JSON-serializable by the default
            serializer.
        topic (str): The message topic as a unicode string. If this is
            not provided, the default topic for the class is used. See the
            attribute documentation below for details.
        properties (pika.BasicProperties): The AMQP properties. If this is not
            provided, they will be generated. Most users should not need to provide
            this, but it can be useful in testing scenarios.
        severity (int): An integer that indicates the severity of the message. This is
            used to determine what messages to notify end users about and should be
            :data:`DEBUG`, :data:`INFO`, :data:`WARNING`, or :data:`ERROR`. The
            default is :data:`INFO`, and can be set as a class attribute or on
            an instance-by-instance basis.

    Attributes:
        id (str): The message id as a unicode string. This attribute is
            automatically generated and set by the library and users should only
            set it themselves in testing scenarios.
        topic (str): The message topic as a unicode string. The topic
            is used by message consumers to filter what messages they receive.
            Topics should be a string of words separated by '.' characters,
            with a length limit of 255 bytes. Because of this byte limit, it is
            best to avoid non-ASCII character. Topics should start general and
            get more specific each word. For example: "bodhi.update.kernel" is
            a possible topic. "bodhi" identifies the application, "update"
            identifies the message, and "kernel" identifies the package in the
            update. This can be set at a class level or on a instance level.
            Dynamic, specific topics that allow for fine-grain filtering are
            preferred.
        headers_schema (dict): A `JSON schema <http://json-schema.org/>`_ to be used with
            :func:`jsonschema.validate` to validate the message headers. For
            most users, the default definition should suffice.
        body_schema (dict): A `JSON schema <http://json-schema.org/>`_ to be used with
            :func:`jsonschema.validate` to validate the message body. The body_schema
            is retrieved on a message instance so it is not required to be a
            class attribute, although this is a convenient approach. Users are
            also free to write the JSON schema as a file and load the file from
            the filesystem or network if they prefer.
        body (dict): The message body as a Python dictionary. This is validated by
            the body schema before publishing and before consuming.
        severity (int): An integer that indicates the severity of the message. This is
            used to determine what messages to notify end users about and should be
            :data:`DEBUG`, :data:`INFO`, :data:`WARNING`, or :data:`ERROR`. The
            default is :data:`INFO`, and can be set as a class attribute or on
            an instance-by-instance basis.
        queue (str): The name of the queue this message arrived through. This
            attribute is set automatically by the library and users should never
            set it themselves.
        deprecated (bool): Whether this message schema has been deprecated by a more
            recent version. Emits a warning when a message of this class is received,
            to let consumers know that they should plan to upgrade. Defaults to
            ``False``.
        priority (int): The priority for the message, if the destination queue
            supports it. Defaults to zero (lowest priority).

            This value is taken into account in queues that have the ``x-max-priority``
            argument set. Most queues in Fedora don't support priorities, in which case
            the value will be ignored.

            Larger numbers indicate higher priority, you can read more about it in
            `RabbitMQ's documentation on priority`_.

            .. _RabbitMQ's documentation on priority:  https://www.rabbitmq.com/priority.html
    """

    severity = INFO
    topic = ""
    headers_schema = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "description": "Schema for message headers",
        "type": "object",
        "properties": {
            "fedora_messaging_severity": {
                "type": "number",
                "enum": [DEBUG, INFO, WARNING, ERROR],
            },
            "fedora_messaging_schema": {"type": "string"},
            "sent-at": {"type": "string"},
        },
    }
    body_schema = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "description": "Schema for message body",
        "type": "object",
    }
    deprecated = False

    def __init__(self, body=None, headers=None, topic=None, properties=None, severity=None):
        self.body = body or {}

        if topic:
            # Default is "" on base class
            self.topic = topic

        headers = headers or {}
        if severity:
            self.severity = severity
        self._properties = properties or self._build_properties(headers)
        self.queue = None

    def _build_properties(self, headers):
        # Consumers use this to determine what schema to use and if they're out
        # of date.
        headers = headers.copy()
        headers["fedora_messaging_schema"] = get_name(self.__class__)
        now = datetime.datetime.now(tz=datetime.timezone.utc).replace(microsecond=0)
        headers["sent-at"] = now.isoformat()
        headers["fedora_messaging_severity"] = self.severity
        # Mirror the priority in the headers for debugging purposes
        headers["priority"] = config.conf["publish_priority"] or 0
        headers.update(self._filter_headers())
        message_id = str(uuid.uuid4())
        return pika.BasicProperties(
            content_type="application/json",
            content_encoding="utf-8",
            delivery_mode=2,
            headers=headers,
            message_id=message_id,
            priority=config.conf["publish_priority"],
        )

    def _filter_headers(self):
        """
        Add headers designed for filtering messages based on objects.

        Returns:
            dict: Filter-related headers to be combined with the existing headers
        """
        headers = {}
        properties = [
            ("user", "usernames"),
            ("group", "groups"),
            ("rpm", "packages"),
            ("container", "containers"),
            ("module", "modules"),
            ("flatpak", "flatpaks"),
        ]
        for header_name, prop_name in properties:
            try:
                items = getattr(self, prop_name)
            except Exception:
                # The message is probably invalid, don't add the header
                _log.warning("Header %s not found in message", prop_name)
                continue
            for item in items:
                headers[f"fedora_messaging_{header_name}_{item}"] = True
        return headers

    @property
    def _headers(self):
        """
        The message headers dictionary.

        .. note: If there's a reason users want to use this interface, it can
              be made public. Please file a bug if you feel you need this.
        """
        return self._properties.headers

    @_headers.setter
    def _headers(self, value):
        self._properties.headers = value

    @property
    def id(self):
        return self._properties.message_id

    @id.setter
    def id(self, value):
        self._properties.message_id = value

    @property
    def priority(self):
        return self._properties.priority or 0

    @priority.setter
    def priority(self, value):
        value = value or 0  # convert None to 0
        self._properties.priority = value
        # Mirror the priority in the headers for debugging purposes
        self._headers["priority"] = value

    @property
    def _encoded_routing_key(self):
        """The encoded routing key used to publish the message on the broker."""
        topic = self.topic
        if config.conf["topic_prefix"]:
            topic = ".".join((config.conf["topic_prefix"].rstrip("."), topic))
        return topic.encode("utf-8")

    @property
    def _encoded_body(self):
        """The encoded body used to publish the message."""
        return json.dumps(self.body).encode("utf-8")

    def __repr__(self):
        """
        Provide a printable representation of the object that can be passed to func:`eval`.
        """
        return (
            f"{self.__class__.__name__}(id={self.id!r}, topic={self.topic!r}, body={self.body!r})"
        )

    def __eq__(self, other):
        """
        Two messages of the same class with the same topic, headers, and body are equal.

        The "sent-at" header is excluded from the equality check as this is set
        automatically and is dependent on when the object is created.

        Args:
            other (object): The object to check for equality.

        Returns:
            bool: True if the messages are equal.
        """
        if not isinstance(other, self.__class__):
            return False

        headers = self._headers.copy()
        other_headers = other._headers.copy()
        try:
            del headers["sent-at"]
        except KeyError:
            pass
        try:
            del other_headers["sent-at"]
        except KeyError:
            pass

        return self.topic == other.topic and self.body == other.body and headers == other_headers

    def validate(self):
        """
        Validate the headers and body with the message schema, if any.

        In addition to the user-provided schema, all messages are checked against
        the base schema which requires certain message headers and the that body
        be a JSON object.

        .. warning:: This method should not be overridden by sub-classes.

        Raises:
            jsonschema.ValidationError: If either the message headers or the message body
                are invalid.
            jsonschema.SchemaError: If either the message header schema or the message body
                schema are invalid.
        """
        for schema in (self.headers_schema, Message.headers_schema):
            _log.debug(
                'Validating message headers "%r" with schema "%r"',
                self._headers,
                schema,
            )
            jsonschema.validate(self._headers, schema)
        for schema in (self.body_schema, Message.body_schema):
            _log.debug('Validating message body "%r" with schema "%r"', self.body, schema)
            jsonschema.validate(self.body, schema)

    @property
    def summary(self):
        """
        A short, human-readable representation of this message.

        This should provide a short summary of the message, much like the subject line
        of an email.

        .. note:: Sub-classes should override this method. It is used to create
            the subject of email notifications, IRC notification, and by other
            tools to display messages to humans in short form.

        The default implementation is to simply return the message topic.
        """
        return self.topic

    def __str__(self):
        """
        A human-readable representation of this message.

        This should provide a detailed, long-form representation of the
        message. The default implementation is to format the raw message id,
        topic, headers, and body.

        .. note:: Sub-classes should override this method. It is used to create
            the body of email notifications and by other tools to display messages
            to humans.
        """
        return "Id: {i}\nTopic: {t}\nHeaders: {h}\nBody: {b}".format(
            i=self.id,
            t=self.topic,
            h=json.dumps(self._headers, sort_keys=True, indent=4, separators=(",", ": ")),
            b=json.dumps(self.body, sort_keys=True, indent=4, separators=(",", ": ")),
        )

    @property
    def url(self):
        """
        An URL to the action that caused this message to be emitted.

        .. note:: Sub-classes should override this method if there is a URL
            associated with message.

        Returns:
            str or None: A relevant URL.
        """
        return None

    @property
    def app_name(self):
        """The name of the application that generated the message.

        .. note:: Sub-classes should override this method.

        Returns:
            str or None: The name of the application.
        """
        return None

    @property
    def app_icon(self):
        """An URL to the icon of the application that generated the message.

        .. note:: Sub-classes should override this method if their application
            has an icon and they wish that image to appear in applications that
            consume messages.

        Returns:
            str or None: The URL to the app's icon.
        """
        return None

    @property
    def agent_name(self):
        """The username of the user who caused the action.

        .. note:: Sub-classes should override this method if the message was
            triggered by a particular user.

        Returns:
            str or None: The agent's username.
        """
        return None

    @property
    def agent_avatar(self):
        """An URL to the avatar of the user who caused the action.

        .. note:: Sub-classes should override this method if the default
            Libravatar and OpenID-based URL generator is not appropriate.

        Returns:
            str or None: The URL to the user's avatar.
        """
        return user_avatar_url(self.agent_name) if self.agent_name is not None else None

    @property
    def usernames(self):
        """List of users affected by the action that generated this message.

        .. note:: Sub-classes should override this method if the message pertains
            to a user or users. The data returned from this property is used to
            filter notifications.

        Returns:
            list(str): A list of affected usernames.
        """
        return []

    @property
    def groups(self):
        """List of groups affected by the action that generated this message.

        .. note:: Sub-classes should override this method if the message pertains
            to a group or groups. The data returned from this property is used to
            filter notifications.

        Returns:
            list(str): A list of affected groups.
        """
        return []

    @property
    def packages(self):
        """List of RPM packages affected by the action that generated this message.

        .. note:: Sub-classes should override this method if the message pertains
            to one or more RPM packages. The data returned from this property
            is used to filter notifications.

        Returns:
            list(str): A list of affected package names.
        """
        return []

    @property
    def containers(self):
        """List of containers affected by the action that generated this message.

        .. note:: Sub-classes should override this method if the message pertains
            to one or more container images. The data returned from this property
            is used to filter notifications.

        Returns:
            list(str): A list of affected container names.
        """
        return []

    @property
    def modules(self):
        """List of modules affected by the action that generated this message.

        .. note:: Sub-classes should override this method if the message pertains
            to one or more modules. The data returned from this property is
            used to filter notifications.

        Returns:
            list(str): A list of affected module names.
        """
        return []

    @property
    def flatpaks(self):
        """List of flatpaks affected by the action that generated this message.

        .. note:: Sub-classes should override this method if the message pertains
            to one or more flatpaks. The data returned from this property is
            used to filter notifications.

        Returns:
            list(str): A list of affected flatpaks names.
        """
        return []


#: The schema for each JSON object produced by :func:`dumps`, consumed by
#: :func:`loads`, and expected by CLI commands like "fedora-messaging publish".
SERIALIZED_MESSAGE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Schema for the JSON object used to represent messages in a file",
    "type": "object",
    "properties": {
        "topic": {"type": "string", "description": "The message topic"},
        "headers": {
            "type": "object",
            "properties": Message.headers_schema["properties"],
            "description": "The message headers",
        },
        "id": {"type": "string", "description": "The message's UUID."},
        "body": {"type": "object", "description": "The message body."},
        "queue": {
            "type": ["string", "null"],
            "description": "The queue the message arrived on, if any.",
        },
        "priority": {
            "type": ["integer", "null"],
            "description": "The priority that the message has been sent with.",
        },
    },
    "required": ["topic", "body"],
}


def load_message(message_dict):
    """Load a message from a dictionary serialization.

    The dictionary must conform to the :data:`SERIALIZED_MESSAGE_SCHEMA` format.

    Args:
        message_dict (dict): The dictionary representing the message.

    Raises:
        ValidationError: If the dictionary does not pass the serialization schema.

    Returns:
        Message: The deserialized message object, as an instance of :class:`Message`
            or one of its subclasses.
    """
    try:
        jsonschema.validate(message_dict, SERIALIZED_MESSAGE_SCHEMA)
    except jsonschema.exceptions.ValidationError as e:
        raise ValidationError(e) from e
    MessageClass = get_class(
        message_dict.get("headers", {}).get("fedora_messaging_schema", "base.message")
    )
    message = MessageClass(
        body=message_dict["body"],
        topic=message_dict["topic"],
        headers=message_dict.get("headers"),
        severity=message_dict.get("headers", {}).get("fedora_messaging_severity"),
    )
    # Restore the sent-at header
    try:
        message._headers["sent-at"] = message_dict["headers"]["sent-at"]
    except KeyError:
        pass
    if "id" in message_dict:
        message.id = message_dict["id"]
    message.queue = message_dict.get("queue")
    message.priority = message_dict.get("priority")
    return message


def dumps(messages):
    """
    Serialize messages to a file format acceptable for :func:`loads` or for the
    publish CLI command. The format is a string where each line is a JSON
    object that conforms to the :data:`SERIALIZED_MESSAGE_SCHEMA` format.

    Args:
        messages (list or Message): The messages to serialize. Each message in
            the messages is subclass of Message.

    Returns:
        str: Serialized messages.

    Raises:
        ValidationError: If one of the messages provided doesn't conform to its
            schema.
    """
    if isinstance(messages, Message):
        messages = [messages]

    serialized_messages = []
    for message in messages:
        try:
            message.validate()
        except (jsonschema.exceptions.ValidationError, AttributeError) as e:
            raise ValidationError(e) from e
        m = {
            "topic": message.topic,
            "headers": message._headers,
            "id": message.id,
            "body": message.body,
            "queue": message.queue,
            "priority": message.priority,
        }
        serialized_messages.append(json.dumps(m, ensure_ascii=False, sort_keys=True))

    return "\n".join(serialized_messages) + "\n"


def loads(serialized_messages):
    """
    Deserialize messages from a file format produced by :func:`dumps`.  The
    format is a string where each line is a JSON object that conforms to the
    :data:`SERIALIZED_MESSAGE_SCHEMA` format.

    Args:
        serialized_messages (str): A string made up of a JSON object per line.

    Returns:
        list: Deserialized message objects.

    Raises:
        ValidationError: If the string isn't formatted properly or message
            doesn't pass the message schema validation
    """
    messages = []
    for serialized_message in serialized_messages.splitlines():
        try:
            message_dict = json.loads(serialized_message)
        except ValueError as e:
            raise ValidationError(e) from e
        message = load_message(message_dict)
        messages.append(message)

    return messages



# This file is part of fedora_messaging.
# Copyright (C) 2018 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Tests for the :module:`fedora_messaging.api` module."""

import logging
from unittest import mock

import pytest
import pytest_twisted
from twisted.internet import defer, threads

from fedora_messaging import api, config
from fedora_messaging.exceptions import PublishException
from fedora_messaging.signals import (
    pre_publish_signal,
    publish_failed_signal,
    publish_signal,
)
from fedora_messaging.twisted import consumer


class TestCheckCallback:
    """Tests for :func:`api._check_callback`"""

    def test_method(self):
        """Assert methods are valid."""

        class Callback:
            def callback(self):
                return

        class_instance = Callback()

        callback = api._check_callback(class_instance.callback)

        assert callback == class_instance.callback

    def test_func(self):
        """Assert functions are valid."""

        def callback(message):
            return

        cb = api._check_callback(callback)

        assert cb is callback

    def test_class(self):
        """Assert classes are instantiated."""

        class Callback:
            def __call__(self, message):
                return "It worked"

        callback = api._check_callback(Callback)

        assert callback(None) == "It worked"

    def test_class_no_call(self):
        """Assert classes are instantiated."""

        class Callback:
            pass

        with pytest.raises(ValueError):
            api._check_callback(Callback)

    def test_class_init_args(self):
        """Assert classes are instantiated."""

        class Callback:
            def __init__(self, args):
                self.args = args

        with pytest.raises(TypeError):
            api._check_callback(Callback)

    def test_not_callable(self):
        """Assert a ValueError is raised for things that can't be called."""
        with pytest.raises(ValueError):
            api._check_callback("")


@mock.patch("fedora_messaging.api._twisted_service")
class TestTwistedConsume:
    """Tests for :func:`api.twisted_consume`"""

    def dummy_callback(self):
        pass

    def test_wrap_bindings(self, mock_service):
        """Assert bindings are always passed to the factory as a list."""
        bindings = {"queue": "q1", "exchange": "e1", "routing_keys": ["#"]}

        def callback(msg):
            pass

        api.twisted_consume(callback, bindings, {})

        mock_service._service.factory.consume.assert_called_once_with(callback, [bindings], {})

    def test_defaults(self, mock_service):
        """Assert that bindings and queues come from the config if not provided."""

        api.twisted_consume(self.dummy_callback)

        mock_service._service.factory.consume.assert_called_once_with(
            self.dummy_callback,
            config.conf["bindings"],
            config.conf["queues"],
        )

    def test_bindings_dict(self, mock_service):
        """Assert consume wraps bindings in a list of just a plain dict"""
        bindings = {"queue": "q1", "exchange": "e1", "routing_keys": ["#"]}

        api.twisted_consume(self.dummy_callback, bindings)

        mock_service._service.factory.consume.assert_called_once_with(
            self.dummy_callback, [bindings], config.conf["queues"]
        )

    def test_bindings_invalid_type(self, mock_service):
        """Assert bindings are validated and result in a value error if they are invalid."""
        with pytest.raises(ValueError):
            api.twisted_consume(self.dummy_callback, "test_bindings")

    def test_bindings_list_of_dict(self, mock_service):
        """Assert consume is working(bindings type is dict)"""
        bindings = [{"queue": "q1", "exchange": "e1", "routing_keys": ["#"]}]

        api.twisted_consume(self.dummy_callback, bindings)

        mock_service._service.factory.consume.assert_called_once_with(
            self.dummy_callback, bindings, config.conf["queues"]
        )

    def test_queues_invalid_type(self, mock_service):
        """Assert queues are validated and result in a value error if they are invalid."""
        with pytest.raises(ValueError):
            api.twisted_consume(self.dummy_callback, None, "should be a dict")

    def test_with_queues(self, mock_service):
        """Assert queues is used over the config if provided."""
        queues = {
            "q1": {
                "durable": True,
                "auto_delete": False,
                "exclusive": False,
                "arguments": {},
            }
        }

        api.twisted_consume(self.dummy_callback, bindings=[], queues=queues)
        mock_service._service.factory.consume.assert_called_once_with(
            self.dummy_callback, [], queues
        )


class TestConsume:
    @mock.patch("fedora_messaging.api.crochet")
    @mock.patch("fedora_messaging.api._twisted_consume_wrapper")
    def test_consume(self, mock_wrapper, mock_crochet):
        """Assert the consume call forwards to the twisted wrapper."""
        api.consume(1, 2, 3)

        mock_crochet.setup.assert_called_once_with()
        mock_wrapper.assert_called_once_with(1, 2, 3)


@mock.patch("fedora_messaging.api._twisted_publish_wrapper")
class TestPublish:

    def test_publish_to_exchange(self, mock_twisted_publish):
        """Assert a message can be published to the exchange."""
        message = "test_message"
        exchange = "test_exchange"

        api.publish(message, exchange)

        mock_twisted_publish.assert_called_once_with(message, exchange)
        mock_twisted_publish.return_value.wait.assert_called_once_with(timeout=30)

    def test_publish_failed(self, mock_twisted_publish):
        """Assert an exception is raised when message can't be published."""
        message = "test_message"
        exchange = "test_exchange"
        expected_exception = PublishException(reason="Unable to publish message")
        mock_twisted_publish.return_value.wait.side_effect = expected_exception

        with pytest.raises(type(expected_exception)):
            api.publish(message, exchange=exchange)

        mock_twisted_publish.assert_called_once_with(message, exchange)


class TestTwistedPublishWrapper:

    @pytest_twisted.inlineCallbacks
    def test_publish_to_exchange(self):
        """Assert a message can be published to the exchange."""
        message = "test_message"
        exchange = "test_exchange"

        with mock.patch("fedora_messaging.api.twisted_publish") as mock_twisted_publish:
            result = api._twisted_publish_wrapper(message, exchange)
            yield threads.deferToThread(result.wait, timeout=1)

        mock_twisted_publish.assert_called_once_with(message, exchange)

    @pytest_twisted.inlineCallbacks
    def test_publish_failed(self):
        """Assert an exception is raised when message can't be published."""
        message = "test_message"
        exchange = "test_exchange"
        expected_exception = PublishException(reason="Unable to publish message")

        with mock.patch("fedora_messaging.api.twisted_publish") as mock_twisted_publish:
            mock_twisted_publish.side_effect = expected_exception
            with pytest.raises(type(expected_exception)):
                result = api._twisted_publish_wrapper(message, exchange=exchange)
                yield threads.deferToThread(result.wait, timeout=1)

        mock_twisted_publish.assert_called_once_with(message, exchange)

    @pytest_twisted.inlineCallbacks
    def test_publish_cancelled(self, caplog):
        """Assert an exception is raised when message can't be published."""
        message = "test_message"
        exchange = "test_exchange"
        expected_exception = defer.CancelledError("dummy error")
        caplog.set_level(logging.DEBUG)

        with mock.patch("fedora_messaging.api.twisted_publish") as mock_twisted_publish:
            mock_twisted_publish.side_effect = expected_exception
            result = api._twisted_publish_wrapper(message, exchange=exchange)
            yield threads.deferToThread(result.wait, timeout=1)

        mock_twisted_publish.assert_called_once_with(message, exchange)
        log_records = [r.getMessage() for r in caplog.records if r.name == "fedora_messaging.api"]
        assert log_records == ["Canceled publish of 'test_message' to test_exchange due to timeout"]


class TestTwistedPublish:
    def setup_method(self, method):
        self.pre_publish_signal_data = {"called": False, "sender": None, "args": None}
        self.publish_signal_data = {"called": False, "sender": None, "args": None}
        self.publish_failed_signal_data = {
            "called": False,
            "sender": None,
            "args": None,
        }

        @pre_publish_signal.connect
        def pre_publish_signal_handler(sender, **kwargs):
            self.pre_publish_signal_data = {
                "called": True,
                "sender": sender,
                "args": kwargs,
            }

        @publish_signal.connect
        def publish_signal_handler(sender, **kwargs):
            self.publish_signal_data = {
                "called": True,
                "sender": sender,
                "args": kwargs,
            }

        @publish_failed_signal.connect
        def publish_failed_signal_handler(sender, **kwargs):
            self.publish_failed_signal_data = {
                "called": True,
                "sender": sender,
                "args": kwargs,
            }

        self.pre_publish_signal_handler = pre_publish_signal_handler
        self.publish_signal_handler = publish_signal_handler
        self.publish_failed_signal_handler = publish_failed_signal_handler

    def teardown_method(self, method):
        pre_publish_signal.disconnect(self.pre_publish_signal_handler)
        publish_signal.disconnect(self.publish_signal_handler)
        publish_failed_signal.disconnect(self.publish_failed_signal_handler)

    @pytest_twisted.inlineCallbacks
    def test_publish_to_exchange(self):
        """Assert a message can be published to the exchange."""
        message = "test_message"
        exchange = "test_exchange"
        expected_pre_publish_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message},
        }
        expected_publish_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message},
        }
        expected_publish_failed_signal_data = {
            "called": False,
            "sender": None,
            "args": None,
        }

        with mock.patch("fedora_messaging.api._twisted_service") as mock_service:
            yield api.twisted_publish(message, exchange)

        mock_service._service.factory.publish.assert_called_once_with(message, exchange=exchange)
        assert self.pre_publish_signal_data == expected_pre_publish_signal_data
        assert self.publish_signal_data == expected_publish_signal_data
        assert self.publish_failed_signal_data == expected_publish_failed_signal_data

    @pytest_twisted.inlineCallbacks
    def test_publish_failed(self):
        """Assert an exception is raised when message can't be published."""
        message = "test_message"
        exchange = "test_exchange"
        expected_exception = PublishException(reason="Unable to publish message")
        expected_pre_publish_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message},
        }
        expected_publish_signal_data = {"called": False, "sender": None, "args": None}
        expected_publish_failed_signal_data = {
            "called": True,
            "sender": api.publish,
            "args": {"message": message, "reason": expected_exception},
        }

        with mock.patch("fedora_messaging.api._twisted_service") as mock_service:
            mock_service._service.factory.publish.side_effect = expected_exception
            with pytest.raises(type(expected_exception)):
                yield api.twisted_publish(message, exchange=exchange)

        mock_service._service.factory.publish.assert_called_once_with(message, exchange=exchange)
        assert self.pre_publish_signal_data == expected_pre_publish_signal_data
        assert self.publish_signal_data == expected_publish_signal_data
        assert self.publish_failed_signal_data == expected_publish_failed_signal_data

    @pytest_twisted.inlineCallbacks
    def test_publish_to_config_exchange(self):
        """Assert a message can be published to the exchange from config."""
        message = "test_message"
        with mock.patch.dict(config.conf, {"publish_exchange": "test_public_exchange"}):
            with mock.patch("fedora_messaging.api._twisted_service") as mock_service:
                yield api.twisted_publish(message)
        mock_service._service.factory.publish.assert_called_once_with(
            message, exchange="test_public_exchange"
        )


@pytest_twisted.inlineCallbacks
def test_consume_unexpected_crash():
    """Assert unexpected exceptions raised inside the Twisted thread are raised."""
    try:
        with mock.patch("fedora_messaging.api._check_callback") as mock_check:
            mock_check.side_effect = RuntimeError("Beep boop you messed up")
            yield threads.deferToThread(api.consume, None)
        pytest.fail("Expected a RuntimeException to be raised")
    except RuntimeError as e:
        assert e.args[0] == "Beep boop you messed up"


@pytest_twisted.inlineCallbacks
def test_consume_successful_halt():
    """Assert consume halts when all consumer.result deferreds have succeeded."""
    consumers = [consumer.Consumer()]
    consumers[0].result = defer.succeed(None)
    try:
        with mock.patch("fedora_messaging.api.twisted_consume") as mock_consume:
            mock_consume.return_value = defer.succeed(consumers)
            d = threads.deferToThread(api.consume, None)
            consumer._add_timeout(d, 0.1)
            yield d
    except (defer.TimeoutError, defer.CancelledError):
        pytest.fail("Expected the consume call to immediately finish, not time out")



# This file is part of fedora_messaging.
# Copyright (C) 2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""Tests for the ``fedora-messaging`` command-line interface."""
import os
import shutil
import subprocess
import sys
import uuid

import pytest
import pytest_twisted
import requests
from twisted.internet import threads

from fedora_messaging import api, exceptions, message

from .utils import RABBITMQ_HOST, sleep


@pytest.fixture
def cli_conf(fixtures_dir):
    return os.path.join(fixtures_dir, "cli_integration.toml")


def halt_exit_0(message):
    """Exit with code 0 when it gets a message"""
    raise exceptions.HaltConsumer()


def halt_exit_42(message):
    """Exit with code 42 when it gets a message"""
    raise exceptions.HaltConsumer(exit_code=42, reason="Life, the universe, and everything")


@pytest.fixture
def queue(scope="function"):
    queue = str(uuid.uuid4())
    yield queue
    requests.delete(
        f"http://{RABBITMQ_HOST}:15672/api/queues/%2F/{queue}",
        auth=("guest", "guest"),
        timeout=3,
    )


@pytest.mark.parametrize(
    "callback,exit_code,msg",
    [
        ("halt_exit_0", 0, b"Consumer indicated it wishes consumption to halt"),
        ("halt_exit_42", 42, b"Life, the universe, and everything"),
    ],
)
@pytest_twisted.inlineCallbacks
def test_consume_halt_with_exitcode(callback, exit_code, msg, queue, cli_conf):
    """Assert user execution halt with reason and exit_code is reported."""
    cmd = shutil.which("fedora-messaging")
    args = [
        sys.executable,
        cmd,
        f"--conf={cli_conf}",
        "consume",
        f"--callback=tests.integration.test_cli:{callback}",
        f"--queue-name={queue}",
        "--exchange=amq.topic",
        "--routing-key=#",
    ]

    process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)  # noqa: S603
    yield sleep(5)

    yield threads.deferToThread(api.publish, message.Message())
    for _ in range(5):
        yield sleep(1)
        if process.poll() is not None:
            break
    else:
        process.kill()
        pytest.fail(f"Process never stopped!: {process.stdout.read()}")

    assert process.returncode == exit_code
    assert msg in process.stdout.read()
